/*
This source is published under the following 3-clause BSD license.

Copyright (c) 2012 - 2013, Lukas Hosek and Alexander Wilkie
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * None of the names of the contributors may be used to endorse or promote 
      products derived from this software without specific prior written 
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


/* ============================================================================

This file is part of a sample implementation of the analytical skylight and
solar radiance models presented in the SIGGRAPH 2012 paper


           "An Analytic Model for Full Spectral Sky-Dome Radiance"

and the 2013 IEEE CG&A paper

       "Adding a Solar Radiance Function to the Hosek Skylight Model"

                                   both by 

                       Lukas Hosek and Alexander Wilkie
                Charles University in Prague, Czech Republic


                        Version: 1.4a, February 22nd, 2013
                        
Version history:

1.4a  February 22nd, 2013
      Removed unnecessary and counter-intuitive solar radius parameters 
      from the interface of the colourspace sky dome initialisation functions.

1.4   February 11th, 2013
      Fixed a bug which caused the relative brightness of the solar disc
      and the sky dome to be off by a factor of about 6. The sun was too 
      bright: this affected both normal and alien sun scenarios. The 
      coefficients of the solar radiance function were changed to fix this.

1.3   January 21st, 2013 (not released to the public)
      Added support for solar discs that are not exactly the same size as
      the terrestrial sun. Also added support for suns with a different
      emission spectrum ("Alien World" functionality).

1.2a  December 18th, 2012
      Fixed a mistake and some inaccuracies in the solar radiance function
      explanations found in ArHosekSkyModel.h. The actual source code is
      unchanged compared to version 1.2.

1.2   December 17th, 2012
      Native RGB data and a solar radiance function that matches the turbidity
      conditions were added.

1.1   September 2012
      The coefficients of the spectral model are now scaled so that the output
      is given in physical units: W / (m^-2 * sr * nm). Also, the output of the
      XYZ model is now no longer scaled to the range [0...1]. Instead, it is
      the result of a simple conversion from spectral data via the CIE 2 degree
      standard observer matching functions. Therefore, after multiplication
      with 683 lm / W, the Y channel now corresponds to luminance in lm.
     
1.0   May 11th, 2012
      Initial release.


Please visit http://cgg.mff.cuni.cz/projects/SkylightModelling/ to check if
an updated version of this code has been published!

============================================================================ */


/*

This file contains the coefficient data for the XYZ colour space version of 
the model.

*/

// Uses Sep 9 pattern / Aug 23 mean dataset

float datasetXYZ1[] = 
{
	// albedo 0, turbidity 1
	-1.117001e+000f,
	-1.867262e-001f,
	-1.113505e+001f,
	1.259865e+001f,
	-3.937339e-002f,
	1.167571e+000f,
	7.100686e-003f,
	3.592678e+000f,
	6.083296e-001f,
	-1.152006e+000f,
	-1.926669e-001f,
	6.152049e+000f,
	-4.770802e+000f,
	-8.704701e-002f,
	7.483626e-001f,
	3.372718e-002f,
	4.464592e+000f,
	4.036546e-001f,
	-1.072371e+000f,
	-2.696632e-001f,
	2.816168e-001f,
	1.820571e+000f,
	-3.742666e-001f,
	2.080607e+000f,
	-7.675295e-002f,
	-2.835366e+000f,
	1.129329e+000f,
	-1.109935e+000f,
	-1.532764e-001f,
	1.198787e+000f,
	-9.015183e-001f,
	5.173015e-003f,
	5.749178e-001f,
	1.075633e-001f,
	4.387949e+000f,
	2.650413e-001f,
	-1.052297e+000f,
	-2.229452e-001f,
	1.952347e+000f,
	5.727205e-001f,
	-4.885070e+000f,
	1.984016e+000f,
	-1.106197e-001f,
	-4.898361e-001f,
	8.907873e-001f,
	-1.070108e+000f,
	-1.600465e-001f,
	1.593886e+000f,
	-4.479251e-005f,
	-3.306541e+000f,
	9.390193e-001f,
	9.513168e-002f,
	2.343583e+000f,
	5.335404e-001f,
	// albedo 0, turbidity 2
	-1.113253e+000f,
	-1.699600e-001f,
	-1.038822e+001f,
	1.137513e+001f,
	-4.040911e-002f,
	1.037455e+000f,
	4.991792e-002f,
	4.801919e+000f,
	6.302710e-001f,
	-1.135747e+000f,
	-1.678594e-001f,
	4.970755e+000f,
	-4.430230e+000f,
	-6.657408e-002f,
	3.636161e-001f,
	1.558009e-001f,
	6.013370e+000f,
	3.959601e-001f,
	-1.095892e+000f,
	-2.732595e-001f,
	7.666496e-001f,
	1.350731e+000f,
	-4.401401e-001f,
	2.470135e+000f,
	-1.707929e-001f,
	-3.260793e+000f,
	1.170337e+000f,
	-1.073668e+000f,
	-2.603929e-002f,
	-1.944589e-001f,
	4.575207e-001f,
	6.878164e-001f,
	-1.390770e-001f,
	3.690299e-001f,
	7.885781e+000f,
	1.877694e-001f,
	-1.070091e+000f,
	-2.798957e-001f,
	2.338478e+000f,
	-2.647221e+000f,
	-7.387808e+000f,
	2.329210e+000f,
	-1.644639e-001f,
	-2.003710e+000f,
	9.874527e-001f,
	-1.067120e+000f,
	-1.418866e-001f,
	1.254090e+000f,
	6.053048e+000f,
	-2.918892e+000f,
	5.322812e-001f,
	1.613053e-001f,
	3.018161e+000f,
	5.274090e-001f,
	// albedo 0, turbidity 3
	-1.129483e+000f,
	-1.890619e-001f,
	-9.065101e+000f,
	9.659923e+000f,
	-3.607819e-002f,
	8.314359e-001f,
	8.181661e-002f,
	4.768868e+000f,
	6.339777e-001f,
	-1.146420e+000f,
	-1.883579e-001f,
	3.309173e+000f,
	-3.127882e+000f,
	-6.938176e-002f,
	3.987113e-001f,
	1.400581e-001f,
	6.283042e+000f,
	5.267076e-001f,
	-1.128348e+000f,
	-2.641305e-001f,
	1.223176e+000f,
	5.514952e-002f,
	-3.490649e-001f,
	1.997784e+000f,
	-4.123709e-002f,
	-2.251251e+000f,
	9.483466e-001f,
	-1.025820e+000f,
	1.404690e-002f,
	-1.187406e+000f,
	2.729900e+000f,
	5.877588e-001f,
	-2.761140e-001f,
	4.602633e-001f,
	8.305125e+000f,
	3.945001e-001f,
	-1.083957e+000f,
	-2.606679e-001f,
	2.207108e+000f,
	-7.202803e+000f,
	-5.968103e+000f,
	2.129455e+000f,
	-7.789512e-002f,
	-1.137688e+000f,
	8.871769e-001f,
	-1.062465e+000f,
	-1.512189e-001f,
	1.042881e+000f,
	1.427839e+001f,
	-4.242214e+000f,
	4.038100e-001f,
	1.997780e-001f,
	2.814449e+000f,
	5.803196e-001f,
	// albedo 0, turbidity 4
	-1.175099e+000f,
	-2.410789e-001f,
	-1.108587e+001f,
	1.133404e+001f,
	-1.819300e-002f,
	6.772942e-001f,
	9.605043e-002f,
	4.231166e+000f,
	6.239972e-001f,
	-1.224207e+000f,
	-2.883527e-001f,
	3.002206e+000f,
	-2.649612e+000f,
	-4.795418e-002f,
	4.984398e-001f,
	3.251434e-002f,
	4.851611e+000f,
	6.551019e-001f,
	-1.136955e+000f,
	-2.423048e-001f,
	1.058823e+000f,
	-2.489236e-001f,
	-2.462179e-001f,
	1.933140e+000f,
	9.106828e-002f,
	-1.905869e-001f,
	8.171065e-001f,
	-1.014535e+000f,
	-8.262500e-003f,
	-1.448017e+000f,
	2.295788e+000f,
	3.510334e-001f,
	-1.477418e+000f,
	5.432449e-001f,
	5.762796e+000f,
	4.908751e-001f,
	-1.070666e+000f,
	-2.379780e-001f,
	1.844589e+000f,
	-5.442448e+000f,
	-4.012768e+000f,
	2.945275e+000f,
	9.854725e-003f,
	8.455959e-002f,
	8.145030e-001f,
	-1.071525e+000f,
	-1.777132e-001f,
	8.076590e-001f,
	9.925865e+000f,
	-3.324623e+000f,
	-6.367437e-001f,
	2.844581e-001f,
	2.248384e+000f,
	6.544022e-001f,
	// albedo 0, turbidity 5
	-1.218818e+000f,
	-2.952382e-001f,
	-1.345975e+001f,
	1.347153e+001f,
	-6.814585e-003f,
	5.079068e-001f,
	1.197230e-001f,
	3.776949e+000f,
	5.836961e-001f,
	-1.409868e+000f,
	-5.114330e-001f,
	2.776539e+000f,
	-2.039001e+000f,
	-2.673769e-002f,
	4.145288e-001f,
	7.829342e-004f,
	2.275883e+000f,
	6.629691e-001f,
	-1.069151e+000f,
	-9.434247e-002f,
	7.293972e-001f,
	-1.222473e+000f,
	-1.533461e-001f,
	2.160357e+000f,
	4.626837e-002f,
	3.852415e+000f,
	8.593570e-001f,
	-1.021306e+000f,
	-1.149551e-001f,
	-1.108414e+000f,
	4.178343e+000f,
	4.013665e-001f,
	-2.222814e+000f,
	6.929462e-001f,
	1.392652e+000f,
	4.401662e-001f,
	-1.074251e+000f,
	-2.224002e-001f,
	1.372356e+000f,
	-8.858704e+000f,
	-3.922660e+000f,
	3.020018e+000f,
	-1.458724e-002f,
	1.511186e+000f,
	8.288064e-001f,
	-1.062048e+000f,
	-1.526582e-001f,
	4.921067e-001f,
	1.485522e+001f,
	-3.229936e+000f,
	-8.426604e-001f,
	3.916243e-001f,
	2.678994e+000f,
	6.689264e-001f,
	// albedo 0, turbidity 6
	-1.257023e+000f,
	-3.364700e-001f,
	-1.527795e+001f,
	1.504223e+001f,
	2.717715e-003f,
	3.029910e-001f,
	1.636851e-001f,
	3.561663e+000f,
	5.283161e-001f,
	-1.635124e+000f,
	-7.329993e-001f,
	3.523939e+000f,
	-2.566337e+000f,
	-1.902543e-002f,
	5.505483e-001f,
	-6.242176e-002f,
	1.065992e+000f,
	6.654236e-001f,
	-9.295823e-001f,
	4.845834e-002f,
	-2.992990e-001f,
	-2.001327e-001f,
	-8.019339e-002f,
	1.807806e+000f,
	9.020277e-002f,
	5.095372e+000f,
	8.639936e-001f,
	-1.093740e+000f,
	-2.148608e-001f,
	-5.216240e-001f,
	2.119777e+000f,
	9.506454e-002f,
	-1.831439e+000f,
	6.961204e-001f,
	1.102084e-001f,
	4.384319e-001f,
	-1.044181e+000f,
	-1.849257e-001f,
	9.071246e-001f,
	-4.648901e+000f,
	-2.279385e+000f,
	2.356502e+000f,
	-4.169147e-002f,
	1.932557e+000f,
	8.296550e-001f,
	-1.061451e+000f,
	-1.458745e-001f,
	2.952267e-001f,
	8.967214e+000f,
	-3.726228e+000f,
	-5.022316e-001f,
	5.684877e-001f,
	3.102347e+000f,
	6.658443e-001f,
	// albedo 0, turbidity 7
	-1.332391e+000f,
	-4.127769e-001f,
	-9.328643e+000f,
	9.046194e+000f,
	3.457775e-003f,
	3.377425e-001f,
	1.530909e-001f,
	3.301209e+000f,
	4.997917e-001f,
	-1.932002e+000f,
	-9.947777e-001f,
	-2.042329e+000f,
	3.586940e+000f,
	-5.642182e-002f,
	8.130478e-001f,
	-8.195988e-002f,
	1.118294e-001f,
	5.617231e-001f,
	-8.707374e-001f,
	1.286999e-001f,
	1.820054e+000f,
	-4.674706e+000f,
	3.317471e-003f,
	5.919018e-001f,
	1.975278e-001f,
	6.686519e+000f,
	9.631727e-001f,
	-1.070378e+000f,
	-3.030579e-001f,
	-9.041938e-001f,
	6.200201e+000f,
	1.232207e-001f,
	-3.650628e-001f,
	5.029403e-001f,
	-2.903162e+000f,
	3.811408e-001f,
	-1.063035e+000f,
	-1.637545e-001f,
	5.853072e-001f,
	-7.889906e+000f,
	-1.200641e+000f,
	1.035018e+000f,
	1.192093e-001f,
	3.267054e+000f,
	8.416151e-001f,
	-1.053655e+000f,
	-1.562286e-001f,
	2.423683e-001f,
	1.128575e+001f,
	-4.363262e+000f,
	-7.314160e-002f,
	5.642088e-001f,
	2.514023e+000f,
	6.670457e-001f,
	// albedo 0, turbidity 8
	-1.366112e+000f,
	-4.718287e-001f,
	-7.876222e+000f,
	7.746900e+000f,
	-9.182309e-003f,
	4.716076e-001f,
	8.320252e-002f,
	3.165603e+000f,
	5.392334e-001f,
	-2.468204e+000f,
	-1.336340e+000f,
	-5.386723e+000f,
	7.072672e+000f,
	-8.329266e-002f,
	8.636876e-001f,
	-1.978177e-002f,
	-1.326218e-001f,
	2.979222e-001f,
	-9.653522e-001f,
	-2.373416e-002f,
	1.810250e+000f,
	-6.467262e+000f,
	1.410706e-001f,
	-4.753717e-001f,
	3.003095e-001f,
	6.551163e+000f,
	1.151083e+000f,
	-8.943186e-001f,
	-2.487152e-001f,
	-2.308960e-001f,
	8.512648e+000f,
	1.298402e-001f,
	1.034705e+000f,
	2.303509e-001f,
	-3.924095e+000f,
	2.982717e-001f,
	-1.146999e+000f,
	-2.318784e-001f,
	8.992419e-002f,
	-9.933614e+000f,
	-8.860920e-001f,
	-3.071656e-002f,
	2.852012e-001f,
	3.046199e+000f,
	8.599001e-001f,
	-1.032399e+000f,
	-1.645145e-001f,
	2.683599e-001f,
	1.327701e+001f,
	-4.407670e+000f,
	7.709869e-002f,
	4.951727e-001f,
	1.957277e+000f,
	6.630943e-001f,
	// albedo 0, turbidity 9
	-1.469070e+000f,
	-6.135092e-001f,
	-6.506263e+000f,
	6.661315e+000f,
	-3.835383e-002f,
	7.150413e-001f,
	7.784318e-003f,
	2.820577e+000f,
	6.756784e-001f,
	-2.501583e+000f,
	-1.247404e+000f,
	-1.523462e+001f,
	1.633191e+001f,
	-1.204803e-002f,
	5.896471e-001f,
	-2.002023e-002f,
	1.144647e+000f,
	6.177874e-002f,
	-2.438672e+000f,
	-1.127291e+000f,
	5.731172e+000f,
	-1.021350e+001f,
	6.165610e-002f,
	-7.752641e-001f,
	4.708254e-001f,
	4.176847e+000f,
	1.200881e+000f,
	-1.513427e-001f,
	9.792731e-002f,
	-1.612349e+000f,
	9.814289e+000f,
	5.188921e-002f,
	1.716403e+000f,
	-7.039255e-002f,
	-2.815115e+000f,
	3.291874e-001f,
	-1.318511e+000f,
	-3.650554e-001f,
	4.221268e-001f,
	-9.294529e+000f,
	-4.397520e-002f,
	-8.100625e-001f,
	3.742719e-001f,
	1.834166e+000f,
	8.223450e-001f,
	-1.016009e+000f,
	-1.820264e-001f,
	1.278426e-001f,
	1.182696e+001f,
	-4.801528e+000f,
	4.947899e-001f,
	4.660378e-001f,
	1.601254e+000f,
	6.702359e-001f,
	// albedo 0, turbidity 10
	-1.841310e+000f,
	-9.781779e-001f,
	-4.610903e+000f,
	4.824662e+000f,
	-5.100806e-002f,
	6.463776e-001f,
	-6.377724e-006f,
	2.216875e+000f,
	8.618530e-001f,
	-2.376373e+000f,
	-1.108657e+000f,
	-1.489799e+001f,
	1.546458e+001f,
	4.091025e-002f,
	9.761780e-002f,
	-1.048958e-002f,
	2.165834e+000f,
	-1.609171e-001f,
	-4.710318e+000f,
	-2.261963e+000f,
	6.947327e+000f,
	-1.034828e+001f,
	-1.325542e-001f,
	7.508674e-001f,
	2.247553e-001f,
	2.873142e+000f,
	1.297100e+000f,
	2.163750e-001f,
	-1.944345e-001f,
	-2.437860e+000f,
	1.011314e+001f,
	4.450500e-001f,
	3.111492e-001f,
	2.751323e-001f,
	-1.627906e+000f,
	2.531213e-001f,
	-1.258794e+000f,
	-3.524641e-001f,
	8.425444e-001f,
	-1.085313e+001f,
	-1.154381e+000f,
	-4.638014e-001f,
	-2.781115e-003f,
	4.344498e-001f,
	8.507091e-001f,
	-1.018938e+000f,
	-1.804153e-001f,
	-6.354054e-002f,
	1.573150e+001f,
	-4.386999e+000f,
	6.211115e-001f,
	5.294648e-001f,
	1.580749e+000f,
	6.586655e-001f,
	// albedo 1, turbidity 1
	-1.116416e+000f,
	-1.917524e-001f,
	-1.068233e+001f,
	1.222221e+001f,
	-3.668978e-002f,
	1.054022e+000f,
	1.592132e-002f,
	3.180583e+000f,
	5.627370e-001f,
	-1.132341e+000f,
	-1.671286e-001f,
	5.976499e+000f,
	-4.227366e+000f,
	-9.542489e-002f,
	8.664938e-001f,
	8.351793e-003f,
	4.876068e+000f,
	4.492779e-001f,
	-1.087635e+000f,
	-3.173679e-001f,
	4.314407e-001f,
	1.100555e+000f,
	-4.410057e-001f,
	1.677253e+000f,
	-3.005925e-002f,
	-4.201249e+000f,
	1.070902e+000f,
	-1.083031e+000f,
	-8.847705e-002f,
	1.291773e+000f,
	4.546776e-001f,
	3.091894e-001f,
	7.261760e-001f,
	4.203659e-002f,
	5.990615e+000f,
	3.704756e-001f,
	-1.057899e+000f,
	-2.246706e-001f,
	2.329563e+000f,
	-1.219656e+000f,
	-5.335260e+000f,
	8.545378e-001f,
	-3.906209e-002f,
	-9.025499e-001f,
	7.797348e-001f,
	-1.073305e+000f,
	-1.522553e-001f,
	1.767063e+000f,
	1.904280e+000f,
	-3.101673e+000f,
	3.995856e-001f,
	2.905192e-002f,
	2.563977e+000f,
	5.753067e-001f,
	// albedo 1, turbidity 2
	-1.113674e+000f,
	-1.759694e-001f,
	-9.754125e+000f,
	1.087391e+001f,
	-3.841093e-002f,
	9.524272e-001f,
	5.680219e-002f,
	4.227034e+000f,
	6.029571e-001f,
	-1.126496e+000f,
	-1.680281e-001f,
	5.332352e+000f,
	-4.575579e+000f,
	-6.761755e-002f,
	3.295335e-001f,
	1.194896e-001f,
	5.570901e+000f,
	4.536185e-001f,
	-1.103074e+000f,
	-2.681801e-001f,
	6.571479e-002f,
	2.396522e+000f,
	-4.551280e-001f,
	2.466331e+000f,
	-1.232022e-001f,
	-3.023201e+000f,
	1.086379e+000f,
	-1.053299e+000f,
	-2.697173e-002f,
	8.379121e-001f,
	-9.681458e-001f,
	5.890692e-001f,
	-4.872027e-001f,
	2.936929e-001f,
	7.510139e+000f,
	3.079122e-001f,
	-1.079553e+000f,
	-2.710448e-001f,
	2.462379e+000f,
	-3.713554e-001f,
	-8.534512e+000f,
	1.828242e+000f,
	-1.686398e-001f,
	-1.961340e+000f,
	8.941077e-001f,
	-1.069741e+000f,
	-1.396394e-001f,
	1.657868e+000f,
	3.236313e+000f,
	-2.706344e+000f,
	-2.948122e-001f,
	1.314816e-001f,
	2.868457e+000f,
	5.413403e-001f,
	// albedo 1, turbidity 3
	-1.131649e+000f,
	-1.954455e-001f,
	-7.751595e+000f,
	8.685861e+000f,
	-4.910871e-002f,
	8.992952e-001f,
	4.710143e-002f,
	4.254818e+000f,
	6.821116e-001f,
	-1.156689e+000f,
	-1.884324e-001f,
	3.163519e+000f,
	-3.091522e+000f,
	-6.613927e-002f,
	-2.575883e-002f,
	1.640065e-001f,
	6.073643e+000f,
	4.453468e-001f,
	-1.079224e+000f,
	-2.621389e-001f,
	9.446437e-001f,
	1.448479e+000f,
	-3.969384e-001f,
	2.626638e+000f,
	-8.101186e-002f,
	-3.016355e+000f,
	1.076295e+000f,
	-1.080832e+000f,
	1.033057e-002f,
	-3.500156e-001f,
	-3.281419e-002f,
	5.655512e-001f,
	-1.156742e+000f,
	4.534710e-001f,
	8.774122e+000f,
	2.772869e-001f,
	-1.051202e+000f,
	-2.679975e-001f,
	2.719109e+000f,
	-2.190316e+000f,
	-6.878798e+000f,
	2.250481e+000f,
	-2.030252e-001f,
	-2.026527e+000f,
	9.701096e-001f,
	-1.089849e+000f,
	-1.598589e-001f,
	1.564748e+000f,
	6.869187e+000f,
	-3.053670e+000f,
	-6.110435e-001f,
	1.644472e-001f,
	2.370452e+000f,
	5.511770e-001f,
	// albedo 1, turbidity 4
	-1.171419e+000f,
	-2.429746e-001f,
	-8.991334e+000f,
	9.571216e+000f,
	-2.772861e-002f,
	6.688262e-001f,
	7.683478e-002f,
	3.785611e+000f,
	6.347635e-001f,
	-1.228554e+000f,
	-2.917562e-001f,
	2.753986e+000f,
	-2.491780e+000f,
	-4.663434e-002f,
	3.118303e-001f,
	7.546506e-002f,
	4.463096e+000f,
	5.955071e-001f,
	-1.093124e+000f,
	-2.447767e-001f,
	9.097406e-001f,
	5.448296e-001f,
	-2.957824e-001f,
	2.024167e+000f,
	-5.152333e-004f,
	-1.069081e+000f,
	9.369565e-001f,
	-1.056994e+000f,
	1.569507e-002f,
	-8.217491e-001f,
	1.870818e+000f,
	7.061930e-001f,
	-1.483928e+000f,
	5.978206e-001f,
	6.864902e+000f,
	3.673332e-001f,
	-1.054871e+000f,
	-2.758129e-001f,
	2.712807e+000f,
	-5.950110e+000f,
	-6.554039e+000f,
	2.447523e+000f,
	-1.895171e-001f,
	-1.454292e+000f,
	9.131738e-001f,
	-1.100218e+000f,
	-1.746241e-001f,
	1.438505e+000f,
	1.115481e+001f,
	-3.266076e+000f,
	-8.837357e-001f,
	1.970100e-001f,
	1.991595e+000f,
	5.907821e-001f,
	// albedo 1, turbidity 5
	-1.207267e+000f,
	-2.913610e-001f,
	-1.103767e+001f,
	1.140724e+001f,
	-1.416800e-002f,
	5.564047e-001f,
	8.476262e-002f,
	3.371255e+000f,
	6.221335e-001f,
	-1.429698e+000f,
	-5.374218e-001f,
	2.837524e+000f,
	-2.221936e+000f,
	-2.422337e-002f,
	9.313758e-002f,
	7.190250e-002f,
	1.869022e+000f,
	5.609035e-001f,
	-1.002274e+000f,
	-6.972810e-002f,
	4.031308e-001f,
	-3.932997e-001f,
	-1.521923e-001f,
	2.390646e+000f,
	-6.893990e-002f,
	2.999661e+000f,
	1.017843e+000f,
	-1.081168e+000f,
	-1.178666e-001f,
	-4.968080e-001f,
	3.919299e+000f,
	6.046866e-001f,
	-2.440615e+000f,
	7.891538e-001f,
	2.140835e+000f,
	2.740470e-001f,
	-1.050727e+000f,
	-2.307688e-001f,
	2.276396e+000f,
	-9.454407e+000f,
	-5.505176e+000f,
	2.992620e+000f,
	-2.450942e-001f,
	6.078372e-001f,
	9.606765e-001f,
	-1.103752e+000f,
	-1.810202e-001f,
	1.375044e+000f,
	1.589095e+001f,
	-3.438954e+000f,
	-1.265669e+000f,
	2.475172e-001f,
	1.680768e+000f,
	5.978056e-001f,
	// albedo 1, turbidity 6
	-1.244324e+000f,
	-3.378542e-001f,
	-1.111001e+001f,
	1.137784e+001f,
	-7.896794e-003f,
	4.808023e-001f,
	9.249904e-002f,
	3.025816e+000f,
	5.880239e-001f,
	-1.593165e+000f,
	-7.027621e-001f,
	2.220896e+000f,
	-1.437709e+000f,
	-1.534738e-002f,
	6.286958e-002f,
	6.644555e-002f,
	1.091727e+000f,
	5.470080e-001f,
	-9.136506e-001f,
	1.344874e-002f,
	7.772636e-001f,
	-1.209396e+000f,
	-1.408978e-001f,
	2.433718e+000f,
	-1.041938e-001f,
	3.791244e+000f,
	1.037916e+000f,
	-1.134968e+000f,
	-1.803315e-001f,
	-9.267335e-001f,
	4.576670e+000f,
	6.851928e-001f,
	-2.805000e+000f,
	8.687208e-001f,
	1.161483e+000f,
	2.571688e-001f,
	-1.017037e+000f,
	-2.053943e-001f,
	2.361640e+000f,
	-9.887818e+000f,
	-5.122889e+000f,
	3.287088e+000f,
	-2.594102e-001f,
	8.578927e-001f,
	9.592340e-001f,
	-1.118723e+000f,
	-1.934942e-001f,
	1.226023e+000f,
	1.674140e+001f,
	-3.277335e+000f,
	-1.629809e+000f,
	2.765232e-001f,
	1.637713e+000f,
	6.113963e-001f,
	// albedo 1, turbidity 7
	-1.314779e+000f,
	-4.119915e-001f,
	-1.241150e+001f,
	1.241578e+001f,
	2.344284e-003f,
	2.980837e-001f,
	1.414613e-001f,
	2.781731e+000f,
	4.998556e-001f,
	-1.926199e+000f,
	-1.020038e+000f,
	2.569200e+000f,
	-1.081159e+000f,
	-2.266833e-002f,
	3.588668e-001f,
	8.750078e-003f,
	-2.452171e-001f,
	4.796758e-001f,
	-7.780002e-001f,
	1.850647e-001f,
	4.445456e-002f,
	-2.409297e+000f,
	-7.816346e-002f,
	1.546790e+000f,
	-2.807227e-002f,
	5.998176e+000f,
	1.132396e+000f,
	-1.179326e+000f,
	-3.578330e-001f,
	-2.392933e-001f,
	6.467883e+000f,
	5.904596e-001f,
	-1.869975e+000f,
	8.045839e-001f,
	-2.498121e+000f,
	1.610633e-001f,
	-1.009956e+000f,
	-1.311896e-001f,
	1.726577e+000f,
	-1.219356e+001f,
	-3.466239e+000f,
	2.343602e+000f,
	-2.252205e-001f,
	2.573681e+000f,
	1.027109e+000f,
	-1.112460e+000f,
	-2.063093e-001f,
	1.233051e+000f,
	2.058946e+001f,
	-4.578074e+000f,
	-1.145643e+000f,
	3.160192e-001f,
	1.420159e+000f,
	5.860212e-001f,
	// albedo 1, turbidity 8
	-1.371689e+000f,
	-4.914196e-001f,
	-1.076610e+001f,
	1.107405e+001f,
	-1.485077e-002f,
	5.936218e-001f,
	3.685482e-002f,
	2.599968e+000f,
	6.002204e-001f,
	-2.436997e+000f,
	-1.377939e+000f,
	2.130141e-002f,
	1.079593e+000f,
	-1.796232e-002f,
	-3.933248e-002f,
	1.610711e-001f,
	-6.901181e-001f,
	1.206416e-001f,
	-8.743368e-001f,
	7.331370e-002f,
	8.734259e-001f,
	-3.743126e+000f,
	-3.151167e-002f,
	1.297596e+000f,
	-7.634926e-002f,
	6.532873e+000f,
	1.435737e+000f,
	-9.810197e-001f,
	-3.521634e-001f,
	-2.855205e-001f,
	7.134674e+000f,
	6.839748e-001f,
	-1.394841e+000f,
	6.952036e-001f,
	-4.633104e+000f,
	-2.173401e-002f,
	-1.122958e+000f,
	-1.691536e-001f,
	1.382360e+000f,
	-1.102913e+001f,
	-2.608171e+000f,
	1.865111e+000f,
	-1.345154e-001f,
	3.112342e+000f,
	1.094134e+000f,
	-1.075586e+000f,
	-2.077415e-001f,
	1.171477e+000f,
	1.793270e+001f,
	-4.656858e+000f,
	-1.036839e+000f,
	3.338295e-001f,
	1.042793e+000f,
	5.739374e-001f,
	// albedo 1, turbidity 9
	-1.465871e+000f,
	-6.364486e-001f,
	-8.833718e+000f,
	9.343650e+000f,
	-3.223600e-002f,
	7.552848e-001f,
	-3.121341e-006f,
	2.249164e+000f,
	8.094662e-001f,
	-2.448924e+000f,
	-1.270878e+000f,
	-4.823703e+000f,
	5.853058e+000f,
	-2.149127e-002f,
	3.581132e-002f,
	-1.230276e-003f,
	4.892553e-001f,
	-1.597657e-001f,
	-2.419809e+000f,
	-1.071337e+000f,
	1.575648e+000f,
	-4.983580e+000f,
	9.545185e-003f,
	5.032615e-001f,
	4.186266e-001f,
	4.634147e+000f,
	1.433517e+000f,
	-1.383278e-001f,
	-2.797095e-002f,
	-1.943067e-001f,
	6.679623e+000f,
	4.118280e-001f,
	-2.744289e-001f,
	-2.118722e-002f,
	-4.337025e+000f,
	1.505072e-001f,
	-1.341872e+000f,
	-2.518572e-001f,
	1.027009e+000f,
	-6.527103e+000f,
	-1.081271e+000f,
	1.015465e+000f,
	2.845789e-001f,
	2.470371e+000f,
	9.278120e-001f,
	-1.040640e+000f,
	-2.367454e-001f,
	1.100744e+000f,
	8.827253e+000f,
	-4.560794e+000f,
	-7.287017e-001f,
	2.842503e-001f,
	6.336593e-001f,
	6.327335e-001f,
	// albedo 1, turbidity 10
	-1.877993e+000f,
	-1.025135e+000f,
	-4.311037e+000f,
	4.715016e+000f,
	-4.711631e-002f,
	6.335844e-001f,
	-7.665398e-006f,
	1.788017e+000f,
	9.001409e-001f,
	-2.281540e+000f,
	-1.137668e+000f,
	-1.036869e+001f,
	1.136254e+001f,
	1.961739e-002f,
	-9.836174e-002f,
	-6.734567e-003f,
	1.320918e+000f,
	-2.400807e-001f,
	-4.904054e+000f,
	-2.315781e+000f,
	5.735999e+000f,
	-8.626257e+000f,
	-1.255643e-001f,
	1.545446e+000f,
	1.396860e-001f,
	2.972897e+000f,
	1.429934e+000f,
	4.077067e-001f,
	-1.833688e-001f,
	-2.450939e+000f,
	9.119433e+000f,
	4.505361e-001f,
	-1.340828e+000f,
	3.973690e-001f,
	-1.785370e+000f,
	9.628711e-002f,
	-1.296052e+000f,
	-3.250526e-001f,
	1.813294e+000f,
	-1.031485e+001f,
	-1.388690e+000f,
	1.239733e+000f,
	-8.989196e-002f,
	-3.389637e-001f,
	9.639560e-001f,
	-1.062181e+000f,
	-2.423444e-001f,
	7.577592e-001f,
	1.566938e+001f,
	-4.462264e+000f,
	-5.742810e-001f,
	3.262259e-001f,
	9.461672e-001f,
	6.232887e-001f,
};

float datasetXYZRad1[] = 
{
	// albedo 0, turbidity 1
	1.560219e+000f,
	1.417388e+000f,
	1.206927e+000f,
	1.091949e+001f,
	5.931416e+000f,
	7.304788e+000f,
	// albedo 0, turbidity 2
	1.533049e+000f,
	1.560532e+000f,
	3.685059e-001f,
	1.355040e+001f,
	5.543711e+000f,
	7.792189e+000f,
	// albedo 0, turbidity 3
	1.471043e+000f,
	1.746088e+000f,
	-9.299697e-001f,
	1.720362e+001f,
	5.473384e+000f,
	8.336416e+000f,
	// albedo 0, turbidity 4
	1.355991e+000f,
	2.109348e+000f,
	-3.295855e+000f,
	2.264843e+001f,
	5.454607e+000f,
	9.304656e+000f,
	// albedo 0, turbidity 5
	1.244963e+000f,
	2.547533e+000f,
	-5.841485e+000f,
	2.756879e+001f,
	5.576104e+000f,
	1.043287e+001f,
	// albedo 0, turbidity 6
	1.175532e+000f,
	2.784634e+000f,
	-7.212225e+000f,
	2.975347e+001f,
	6.472980e+000f,
	1.092331e+001f,
	// albedo 0, turbidity 7
	1.082973e+000f,
	3.118094e+000f,
	-8.934293e+000f,
	3.186879e+001f,
	8.473885e+000f,
	1.174019e+001f,
	// albedo 0, turbidity 8
	9.692500e-001f,
	3.349574e+000f,
	-1.003810e+001f,
	3.147654e+001f,
	1.338931e+001f,
	1.272547e+001f,
	// albedo 0, turbidity 9
	8.547044e-001f,
	3.151538e+000f,
	-9.095567e+000f,
	2.554995e+001f,
	2.273219e+001f,
	1.410398e+001f,
	// albedo 0, turbidity 10
	7.580340e-001f,
	2.311153e+000f,
	-5.170814e+000f,
	1.229669e+001f,
	3.686529e+001f,
	1.598882e+001f,
	// albedo 1, turbidity 1
	1.664273e+000f,
	1.574468e+000f,
	1.422078e+000f,
	9.768247e+000f,
	1.447338e+001f,
	1.644988e+001f,
	// albedo 1, turbidity 2
	1.638295e+000f,
	1.719586e+000f,
	5.786675e-001f,
	1.239846e+001f,
	1.415419e+001f,
	1.728605e+001f,
	// albedo 1, turbidity 3
	1.572623e+000f,
	1.921559e+000f,
	-7.714802e-001f,
	1.609246e+001f,
	1.420954e+001f,
	1.825908e+001f,
	// albedo 1, turbidity 4
	1.468395e+000f,
	2.211970e+000f,
	-2.845869e+000f,
	2.075027e+001f,
	1.524822e+001f,
	1.937622e+001f,
	// albedo 1, turbidity 5
	1.355047e+000f,
	2.556469e+000f,
	-4.960920e+000f,
	2.460237e+001f,
	1.648360e+001f,
	2.065648e+001f,
	// albedo 1, turbidity 6
	1.291642e+000f,
	2.742036e+000f,
	-6.061967e+000f,
	2.602002e+001f,
	1.819144e+001f,
	2.116712e+001f,
	// albedo 1, turbidity 7
	1.194565e+000f,
	2.972120e+000f,
	-7.295779e+000f,
	2.691805e+001f,
	2.124880e+001f,
	2.201819e+001f,
	// albedo 1, turbidity 8
	1.083631e+000f,
	3.047021e+000f,
	-7.766096e+000f,
	2.496261e+001f,
	2.744264e+001f,
	2.291875e+001f,
	// albedo 1, turbidity 9
	9.707994e-001f,
	2.736459e+000f,
	-6.308284e+000f,
	1.760860e+001f,
	3.776291e+001f,
	2.392150e+001f,
	// albedo 1, turbidity 10
	8.574294e-001f,
	1.865155e+000f,
	-2.364707e+000f,
	4.337793e+000f,
	5.092831e+001f,
	2.523432e+001f,
};

float datasetXYZ2[] = 
{
	// albedo 0, turbidity 1
	-1.127942e+000f,
	-1.905548e-001f,
	-1.252356e+001f,
	1.375799e+001f,
	-3.624732e-002f,
	1.055453e+000f,
	1.385036e-002f,
	4.176970e+000f,
	5.928345e-001f,
	-1.155260e+000f,
	-1.778135e-001f,
	6.216056e+000f,
	-5.254116e+000f,
	-8.787445e-002f,
	8.434621e-001f,
	4.025734e-002f,
	6.195322e+000f,
	3.111856e-001f,
	-1.125624e+000f,
	-3.217593e-001f,
	5.043919e-001f,
	1.686284e+000f,
	-3.536071e-001f,
	1.476321e+000f,
	-7.899019e-002f,
	-4.522531e+000f,
	1.271691e+000f,
	-1.081801e+000f,
	-1.033234e-001f,
	9.995550e-001f,
	7.482946e-003f,
	-6.776018e-002f,
	1.463141e+000f,
	9.492021e-002f,
	5.612723e+000f,
	1.298846e-001f,
	-1.075320e+000f,
	-2.402711e-001f,
	2.141284e+000f,
	-1.203359e+000f,
	-4.945188e+000f,
	1.437221e+000f,
	-8.096750e-002f,
	-1.028378e+000f,
	1.004164e+000f,
	-1.073337e+000f,
	-1.516517e-001f,
	1.639379e+000f,
	2.304669e+000f,
	-3.214244e+000f,
	1.286245e+000f,
	5.613957e-002f,
	2.480902e+000f,
	4.999363e-001f,
	// albedo 0, turbidity 2
	-1.128399e+000f,
	-1.857793e-001f,
	-1.089863e+001f,
	1.172984e+001f,
	-3.768099e-002f,
	9.439285e-001f,
	4.869335e-002f,
	4.845114e+000f,
	6.119211e-001f,
	-1.114002e+000f,
	-1.399280e-001f,
	4.963800e+000f,
	-4.685500e+000f,
	-7.780879e-002f,
	4.049736e-001f,
	1.586297e-001f,
	7.770264e+000f,
	3.449006e-001f,
	-1.185472e+000f,
	-3.403543e-001f,
	6.588322e-001f,
	1.133713e+000f,
	-4.118674e-001f,
	2.061191e+000f,
	-1.882768e-001f,
	-4.372586e+000f,
	1.223530e+000f,
	-1.002272e+000f,
	2.000703e-002f,
	7.073269e-002f,
	1.485075e+000f,
	5.005589e-001f,
	4.301494e-001f,
	3.626541e-001f,
	7.921098e+000f,
	1.574766e-001f,
	-1.121006e+000f,
	-3.007777e-001f,
	2.242051e+000f,
	-4.571561e+000f,
	-7.761071e+000f,
	2.053404e+000f,
	-1.524018e-001f,
	-1.886162e+000f,
	1.018208e+000f,
	-1.058864e+000f,
	-1.358673e-001f,
	1.389667e+000f,
	8.633409e+000f,
	-3.437249e+000f,
	7.295429e-001f,
	1.514700e-001f,
	2.842513e+000f,
	5.014325e-001f,
	// albedo 0, turbidity 3
	-1.144464e+000f,
	-2.043799e-001f,
	-1.020188e+001f,
	1.071247e+001f,
	-3.256693e-002f,
	7.860205e-001f,
	6.872719e-002f,
	4.824771e+000f,
	6.259836e-001f,
	-1.170104e+000f,
	-2.118626e-001f,
	4.391405e+000f,
	-4.198900e+000f,
	-7.111559e-002f,
	3.890442e-001f,
	1.024831e-001f,
	6.282535e+000f,
	5.365688e-001f,
	-1.129171e+000f,
	-2.552880e-001f,
	2.238298e-001f,
	7.314295e-001f,
	-3.562730e-001f,
	1.881931e+000f,
	-3.078716e-002f,
	-1.039120e+000f,
	9.096301e-001f,
	-1.042294e+000f,
	4.450203e-003f,
	-5.116033e-001f,
	2.627589e+000f,
	6.098996e-001f,
	-1.264638e-001f,
	4.325281e-001f,
	7.080503e+000f,
	4.583646e-001f,
	-1.082293e+000f,
	-2.723056e-001f,
	2.065076e+000f,
	-8.143133e+000f,
	-7.892212e+000f,
	2.142231e+000f,
	-7.106240e-002f,
	-1.122398e+000f,
	8.338505e-001f,
	-1.071715e+000f,
	-1.426568e-001f,
	1.095351e+000f,
	1.729783e+001f,
	-3.851931e+000f,
	4.360514e-001f,
	2.114440e-001f,
	2.970832e+000f,
	5.944389e-001f,
	// albedo 0, turbidity 4
	-1.195909e+000f,
	-2.590449e-001f,
	-1.191037e+001f,
	1.207947e+001f,
	-1.589842e-002f,
	6.297846e-001f,
	9.054772e-002f,
	4.285959e+000f,
	5.933752e-001f,
	-1.245763e+000f,
	-3.316637e-001f,
	4.293660e+000f,
	-3.694011e+000f,
	-4.699947e-002f,
	4.843684e-001f,
	2.130425e-002f,
	4.097549e+000f,
	6.530809e-001f,
	-1.148742e+000f,
	-1.902509e-001f,
	-2.393233e-001f,
	-2.441254e-001f,
	-2.610918e-001f,
	1.846988e+000f,
	3.532866e-002f,
	2.660106e+000f,
	8.358294e-001f,
	-1.016080e+000f,
	-7.444960e-002f,
	-5.053436e-001f,
	4.388855e+000f,
	6.054987e-001f,
	-1.208300e+000f,
	5.817215e-001f,
	2.543570e+000f,
	4.726568e-001f,
	-1.072027e+000f,
	-2.101440e-001f,
	1.518378e+000f,
	-1.060119e+001f,
	-6.016546e+000f,
	2.649475e+000f,
	-5.166992e-002f,
	1.571269e+000f,
	8.344622e-001f,
	-1.072365e+000f,
	-1.511201e-001f,
	7.478010e-001f,
	1.900732e+001f,
	-3.950387e+000f,
	-3.473907e-001f,
	3.797211e-001f,
	2.782949e+000f,
	6.296808e-001f,
	// albedo 0, turbidity 5
	-1.239423e+000f,
	-3.136289e-001f,
	-1.351100e+001f,
	1.349468e+001f,
	-7.070423e-003f,
	5.012315e-001f,
	1.106008e-001f,
	3.803619e+000f,
	5.577948e-001f,
	-1.452524e+000f,
	-5.676944e-001f,
	2.993153e+000f,
	-2.277288e+000f,
	-2.168954e-002f,
	3.056720e-001f,
	1.152338e-002f,
	1.852697e+000f,
	6.427228e-001f,
	-1.061421e+000f,
	-4.590521e-002f,
	6.057022e-001f,
	-1.096835e+000f,
	-1.504952e-001f,
	2.344921e+000f,
	-5.491832e-002f,
	5.268322e+000f,
	9.082253e-001f,
	-1.042373e+000f,
	-1.769498e-001f,
	-1.075388e+000f,
	3.831712e+000f,
	3.154140e-001f,
	-2.416458e+000f,
	7.909032e-001f,
	-1.492892e-002f,
	3.854049e-001f,
	-1.064159e+000f,
	-1.892684e-001f,
	1.438685e+000f,
	-8.166362e+000f,
	-3.616364e+000f,
	3.275206e+000f,
	-1.203825e-001f,
	2.039491e+000f,
	8.688057e-001f,
	-1.070120e+000f,
	-1.569508e-001f,
	4.124760e-001f,
	1.399683e+001f,
	-3.547085e+000f,
	-1.046326e+000f,
	4.973825e-001f,
	2.791231e+000f,
	6.503286e-001f,
	// albedo 0, turbidity 6
	-1.283579e+000f,
	-3.609518e-001f,
	-1.335397e+001f,
	1.315248e+001f,
	-4.431938e-004f,
	3.769526e-001f,
	1.429824e-001f,
	3.573613e+000f,
	4.998696e-001f,
	-1.657952e+000f,
	-7.627948e-001f,
	1.958222e+000f,
	-7.949816e-001f,
	-2.882837e-002f,
	5.356149e-001f,
	-5.191946e-002f,
	8.869955e-001f,
	6.263320e-001f,
	-9.527600e-001f,
	6.494189e-002f,
	5.361303e-001f,
	-2.129590e+000f,
	-9.258630e-002f,
	1.604776e+000f,
	5.067770e-002f,
	6.376055e+000f,
	9.138052e-001f,
	-1.080827e+000f,
	-2.523120e-001f,
	-7.154262e-001f,
	4.120085e+000f,
	1.878228e-001f,
	-1.492158e+000f,
	6.881655e-001f,
	-1.446611e+000f,
	4.040631e-001f,
	-1.054075e+000f,
	-1.665498e-001f,
	9.191052e-001f,
	-6.636943e+000f,
	-1.894826e+000f,
	2.107810e+000f,
	-3.680499e-002f,
	2.655452e+000f,
	8.413840e-001f,
	-1.061127e+000f,
	-1.448849e-001f,
	2.667493e-001f,
	1.034103e+001f,
	-4.285769e+000f,
	-3.874504e-001f,
	5.998752e-001f,
	3.132426e+000f,
	6.652753e-001f,
	// albedo 0, turbidity 7
	-1.347345e+000f,
	-4.287832e-001f,
	-9.305553e+000f,
	9.133813e+000f,
	-3.173527e-003f,
	3.977564e-001f,
	1.151420e-001f,
	3.320564e+000f,
	4.998134e-001f,
	-1.927296e+000f,
	-9.901372e-001f,
	-2.593499e+000f,
	4.087421e+000f,
	-5.833993e-002f,
	8.158929e-001f,
	-4.681279e-002f,
	2.423716e-001f,
	4.938052e-001f,
	-9.470092e-001f,
	7.325237e-002f,
	2.064735e+000f,
	-5.167540e+000f,
	-1.313751e-002f,
	4.832169e-001f,
	1.126295e-001f,
	6.970522e+000f,
	1.035022e+000f,
	-1.022557e+000f,
	-2.762616e-001f,
	-9.375748e-001f,
	6.696739e+000f,
	2.200765e-001f,
	-1.133253e-001f,
	5.492505e-001f,
	-3.109391e+000f,
	3.321914e-001f,
	-1.087444e+000f,
	-1.836263e-001f,
	6.225024e-001f,
	-8.576765e+000f,
	-1.107637e+000f,
	7.859427e-001f,
	9.910909e-002f,
	3.112938e+000f,
	8.596261e-001f,
	-1.051544e+000f,
	-1.546262e-001f,
	2.371731e-001f,
	1.200502e+001f,
	-4.527291e+000f,
	7.268862e-002f,
	5.571478e-001f,
	2.532873e+000f,
	6.662000e-001f,
	// albedo 0, turbidity 8
	-1.375576e+000f,
	-4.840019e-001f,
	-8.121290e+000f,
	8.058140e+000f,
	-1.445661e-002f,
	5.123314e-001f,
	5.813321e-002f,
	3.203219e+000f,
	5.442318e-001f,
	-2.325221e+000f,
	-1.241463e+000f,
	-7.063430e+000f,
	8.741369e+000f,
	-7.829950e-002f,
	8.844273e-001f,
	-3.471106e-002f,
	1.740583e-001f,
	2.814079e-001f,
	-1.228700e+000f,
	-2.013412e-001f,
	2.949042e+000f,
	-7.371945e+000f,
	1.071753e-001f,
	-2.491970e-001f,
	2.265223e-001f,
	6.391504e+000f,
	1.172389e+000f,
	-7.601786e-001f,
	-1.680631e-001f,
	-7.584444e-001f,
	8.541356e+000f,
	8.222291e-002f,
	6.729633e-001f,
	3.206615e-001f,
	-3.700940e+000f,
	2.710054e-001f,
	-1.191166e+000f,
	-2.672347e-001f,
	2.927498e-001f,
	-9.713613e+000f,
	-4.783721e-001f,
	2.352803e-001f,
	2.161949e-001f,
	2.691481e+000f,
	8.745447e-001f,
	-1.030135e+000f,
	-1.653301e-001f,
	2.263443e-001f,
	1.296157e+001f,
	-4.650644e+000f,
	7.055709e-003f,
	5.091975e-001f,
	2.000370e+000f,
	6.603839e-001f,
	// albedo 0, turbidity 9
	-1.508018e+000f,
	-6.460933e-001f,
	-6.402745e+000f,
	6.545995e+000f,
	-3.750320e-002f,
	6.921803e-001f,
	3.309819e-003f,
	2.797527e+000f,
	6.978446e-001f,
	-2.333308e+000f,
	-1.167837e+000f,
	-1.746787e+001f,
	1.868630e+001f,
	-8.948229e-003f,
	5.621946e-001f,
	-3.402626e-002f,
	1.217943e+000f,
	1.149865e-002f,
	-2.665953e+000f,
	-1.226307e+000f,
	7.169725e+000f,
	-1.159434e+001f,
	3.583420e-002f,
	-3.074378e-001f,
	3.412248e-001f,
	4.422122e+000f,
	1.283791e+000f,
	-9.705116e-002f,
	8.312991e-002f,
	-2.160462e+000f,
	1.028235e+001f,
	3.543357e-002f,
	1.032049e+000f,
	1.058310e-001f,
	-2.972898e+000f,
	2.418628e-001f,
	-1.329617e+000f,
	-3.699557e-001f,
	5.560117e-001f,
	-9.730113e+000f,
	9.938865e-002f,
	-3.071488e-001f,
	2.510691e-001f,
	1.777111e+000f,
	8.705142e-001f,
	-1.019387e+000f,
	-1.893247e-001f,
	1.194079e-001f,
	1.239436e+001f,
	-4.799224e+000f,
	2.940213e-001f,
	4.841268e-001f,
	1.529724e+000f,
	6.582615e-001f,
	// albedo 0, turbidity 10
	-1.896737e+000f,
	-1.005442e+000f,
	-6.411032e+000f,
	6.548220e+000f,
	-3.227596e-002f,
	5.717262e-001f,
	-8.115192e-006f,
	2.296704e+000f,
	9.000749e-001f,
	-2.411116e+000f,
	-1.225587e+000f,
	-1.753629e+001f,
	1.829393e+001f,
	1.247555e-002f,
	2.364616e-001f,
	-5.114637e-003f,
	1.603778e+000f,
	-2.224156e-001f,
	-4.707121e+000f,
	-2.074977e+000f,
	7.942300e+000f,
	-1.132407e+001f,
	-5.415654e-002f,
	5.446811e-001f,
	1.032493e-001f,
	4.010235e+000f,
	1.369802e+000f,
	1.010482e-001f,
	-4.013305e-001f,
	-2.674579e+000f,
	9.779409e+000f,
	1.782506e-001f,
	7.053045e-001f,
	4.200002e-001f,
	-2.400671e+000f,
	1.953165e-001f,
	-1.243526e+000f,
	-3.391255e-001f,
	8.848882e-001f,
	-9.789025e+000f,
	-3.997324e-001f,
	-9.546227e-001f,
	-1.044017e-001f,
	6.010593e-001f,
	8.714462e-001f,
	-1.014633e+000f,
	-1.730009e-001f,
	-7.738934e-002f,
	1.390903e+001f,
	-4.847307e+000f,
	1.076059e+000f,
	5.685743e-001f,
	1.572992e+000f,
	6.561432e-001f,
	// albedo 1, turbidity 1
	-1.122998e+000f,
	-1.881183e-001f,
	-1.030709e+001f,
	1.158932e+001f,
	-4.079495e-002f,
	9.603774e-001f,
	3.079436e-002f,
	4.009235e+000f,
	5.060745e-001f,
	-1.134790e+000f,
	-1.539688e-001f,
	5.478405e+000f,
	-4.217270e+000f,
	-1.043858e-001f,
	7.165008e-001f,
	1.524765e-002f,
	6.473623e+000f,
	4.207882e-001f,
	-1.134957e+000f,
	-3.513318e-001f,
	7.393837e-001f,
	1.354415e+000f,
	-4.764078e-001f,
	1.690441e+000f,
	-5.492640e-002f,
	-5.563523e+000f,
	1.145743e+000f,
	-1.058344e+000f,
	-5.758503e-002f,
	1.168230e+000f,
	3.269824e-001f,
	1.795193e-001f,
	7.849011e-001f,
	7.441853e-002f,
	6.904804e+000f,
	2.818790e-001f,
	-1.075194e+000f,
	-2.355813e-001f,
	2.463685e+000f,
	-1.536505e+000f,
	-7.505771e+000f,
	9.619712e-001f,
	-6.465851e-002f,
	-1.355492e+000f,
	8.489847e-001f,
	-1.079030e+000f,
	-1.465328e-001f,
	1.773838e+000f,
	2.310131e+000f,
	-3.136065e+000f,
	3.507952e-001f,
	4.435014e-002f,
	2.819225e+000f,
	5.689008e-001f,
	// albedo 1, turbidity 2
	-1.125833e+000f,
	-1.870849e-001f,
	-9.555833e+000f,
	1.059713e+001f,
	-4.225402e-002f,
	9.164663e-001f,
	4.338796e-002f,
	4.400980e+000f,
	6.056119e-001f,
	-1.127440e+000f,
	-1.551891e-001f,
	4.755621e+000f,
	-4.408806e+000f,
	-7.851763e-002f,
	2.268284e-001f,
	1.460070e-001f,
	7.048003e+000f,
	3.525997e-001f,
	-1.143788e+000f,
	-3.170178e-001f,
	5.480669e-001f,
	2.041830e+000f,
	-4.532139e-001f,
	2.302233e+000f,
	-1.887419e-001f,
	-4.489221e+000f,
	1.250967e+000f,
	-1.032849e+000f,
	7.376031e-003f,
	5.666073e-001f,
	-2.312203e-001f,
	4.862894e-001f,
	-1.748294e-001f,
	3.572870e-001f,
	8.380522e+000f,
	1.302333e-001f,
	-1.093728e+000f,
	-2.786977e-001f,
	2.641272e+000f,
	-1.507494e+000f,
	-8.731243e+000f,
	1.684055e+000f,
	-2.023377e-001f,
	-2.176398e+000f,
	1.013249e+000f,
	-1.076578e+000f,
	-1.456205e-001f,
	1.693935e+000f,
	2.945003e+000f,
	-2.822673e+000f,
	-2.520033e-001f,
	1.517034e-001f,
	2.649109e+000f,
	5.179094e-001f,
	// albedo 1, turbidity 3
	-1.146417e+000f,
	-2.119353e-001f,
	-7.187525e+000f,
	8.058599e+000f,
	-5.256438e-002f,
	8.375733e-001f,
	3.887093e-002f,
	4.222111e+000f,
	6.695347e-001f,
	-1.173674e+000f,
	-2.067025e-001f,
	2.899359e+000f,
	-2.804918e+000f,
	-8.473899e-002f,
	3.944225e-003f,
	1.340641e-001f,
	6.160887e+000f,
	4.527141e-001f,
	-1.090098e+000f,
	-2.599633e-001f,
	9.180856e-001f,
	1.092710e+000f,
	-4.215019e-001f,
	2.427660e+000f,
	-9.277667e-002f,
	-2.123523e+000f,
	1.058159e+000f,
	-1.084460e+000f,
	8.056181e-003f,
	-2.453510e-001f,
	6.619567e-001f,
	4.668118e-001f,
	-9.526719e-001f,
	4.648454e-001f,
	8.001572e+000f,
	3.054194e-001f,
	-1.053728e+000f,
	-2.765784e-001f,
	2.792388e+000f,
	-3.489517e+000f,
	-8.150535e+000f,
	2.195757e+000f,
	-2.017234e-001f,
	-2.128017e+000f,
	9.326589e-001f,
	-1.099348e+000f,
	-1.593939e-001f,
	1.568292e+000f,
	7.247853e+000f,
	-2.933000e+000f,
	-5.890481e-001f,
	1.724440e-001f,
	2.433484e+000f,
	5.736558e-001f,
	// albedo 1, turbidity 4
	-1.185983e+000f,
	-2.581184e-001f,
	-7.761056e+000f,
	8.317053e+000f,
	-3.351773e-002f,
	6.676667e-001f,
	5.941733e-002f,
	3.820727e+000f,
	6.324032e-001f,
	-1.268591e+000f,
	-3.398067e-001f,
	2.348503e+000f,
	-2.023779e+000f,
	-5.368458e-002f,
	1.083282e-001f,
	8.402858e-002f,
	3.910254e+000f,
	5.577481e-001f,
	-1.071353e+000f,
	-1.992459e-001f,
	7.878387e-001f,
	1.974702e-001f,
	-3.033058e-001f,
	2.335298e+000f,
	-8.205259e-002f,
	7.954454e-001f,
	9.972312e-001f,
	-1.089513e+000f,
	-3.104364e-002f,
	-5.995746e-001f,
	2.330281e+000f,
	6.581939e-001f,
	-1.821467e+000f,
	6.679973e-001f,
	5.090195e+000f,
	3.125161e-001f,
	-1.040214e+000f,
	-2.570934e-001f,
	2.660489e+000f,
	-6.506045e+000f,
	-7.053586e+000f,
	2.763153e+000f,
	-2.433632e-001f,
	-7.648176e-001f,
	9.452937e-001f,
	-1.116052e+000f,
	-1.831993e-001f,
	1.457694e+000f,
	1.163608e+001f,
	-3.216426e+000f,
	-1.045594e+000f,
	2.285002e-001f,
	1.817407e+000f,
	5.810396e-001f,
	// albedo 1, turbidity 5
	-1.230134e+000f,
	-3.136264e-001f,
	-8.909301e+000f,
	9.145006e+000f,
	-1.055387e-002f,
	4.467317e-001f,
	1.016826e-001f,
	3.342964e+000f,
	5.633840e-001f,
	-1.442907e+000f,
	-5.593147e-001f,
	2.156447e+000f,
	-1.241657e+000f,
	-3.512130e-002f,
	3.050274e-001f,
	1.797175e-002f,
	1.742358e+000f,
	5.977153e-001f,
	-1.027627e+000f,
	-6.481539e-002f,
	4.351975e-001f,
	-1.051677e+000f,
	-2.030672e-001f,
	1.942684e+000f,
	-3.615993e-002f,
	4.050266e+000f,
	9.801624e-001f,
	-1.082110e+000f,
	-1.578209e-001f,
	-3.397511e-001f,
	4.163851e+000f,
	6.650368e-001f,
	-1.841730e+000f,
	7.062544e-001f,
	6.789881e-001f,
	3.172623e-001f,
	-1.047447e+000f,
	-1.977560e-001f,
	2.183364e+000f,
	-8.805249e+000f,
	-5.483962e+000f,
	2.551309e+000f,
	-1.779640e-001f,
	1.519501e+000f,
	9.212536e-001f,
	-1.111853e+000f,
	-1.935736e-001f,
	1.394408e+000f,
	1.392405e+001f,
	-3.465430e+000f,
	-1.068432e+000f,
	2.388671e-001f,
	1.455336e+000f,
	6.233425e-001f,
	// albedo 1, turbidity 6
	-1.262238e+000f,
	-3.546341e-001f,
	-1.008703e+001f,
	1.020084e+001f,
	-1.852187e-003f,
	3.537580e-001f,
	1.239199e-001f,
	3.056093e+000f,
	5.132052e-001f,
	-1.613810e+000f,
	-7.355585e-001f,
	2.760123e+000f,
	-1.685253e+000f,
	-2.517552e-002f,
	2.914258e-001f,
	4.743448e-003f,
	8.689596e-001f,
	5.674192e-001f,
	-9.462336e-001f,
	2.950767e-002f,
	-2.613816e-001f,
	-7.398653e-001f,
	-1.315558e-001f,
	1.901042e+000f,
	-6.447844e-002f,
	4.969341e+000f,
	1.027342e+000f,
	-1.111481e+000f,
	-2.194054e-001f,
	-9.004538e-002f,
	3.983442e+000f,
	4.871278e-001f,
	-1.965315e+000f,
	7.956121e-001f,
	-2.363225e-001f,
	2.718037e-001f,
	-1.036397e+000f,
	-1.827106e-001f,
	1.964747e+000f,
	-8.870759e+000f,
	-4.208011e+000f,
	2.461215e+000f,
	-2.158905e-001f,
	1.561676e+000f,
	9.436866e-001f,
	-1.113769e+000f,
	-1.947819e-001f,
	1.300720e+000f,
	1.516476e+001f,
	-4.088732e+000f,
	-1.069384e+000f,
	2.836434e-001f,
	1.671451e+000f,
	6.229612e-001f,
	// albedo 1, turbidity 7
	-1.328069e+000f,
	-4.244047e-001f,
	-8.417040e+000f,
	8.552244e+000f,
	-6.813504e-003f,
	4.127422e-001f,
	9.619897e-002f,
	2.854227e+000f,
	5.059880e-001f,
	-1.927552e+000f,
	-1.025290e+000f,
	9.529576e-001f,
	4.255950e-001f,
	-3.738779e-002f,
	2.584586e-001f,
	4.911004e-002f,
	-2.640913e-001f,
	4.138626e-001f,
	-8.488094e-001f,
	1.435988e-001f,
	6.356807e-001f,
	-2.895732e+000f,
	-8.473961e-002f,
	1.701305e+000f,
	-1.323908e-001f,
	6.499338e+000f,
	1.210928e+000f,
	-1.128313e+000f,
	-3.397048e-001f,
	-4.043140e-001f,
	6.265097e+000f,
	5.482395e-001f,
	-2.057614e+000f,
	8.884087e-001f,
	-2.943879e+000f,
	9.760301e-002f,
	-1.039764e+000f,
	-1.494772e-001f,
	1.781915e+000f,
	-1.153012e+001f,
	-3.379232e+000f,
	2.517231e+000f,
	-2.764393e-001f,
	2.588849e+000f,
	1.052120e+000f,
	-1.108447e+000f,
	-2.012251e-001f,
	1.198640e+000f,
	1.925331e+001f,
	-4.423892e+000f,
	-1.257122e+000f,
	3.395690e-001f,
	1.481220e+000f,
	5.880175e-001f,
	// albedo 1, turbidity 8
	-1.374185e+000f,
	-4.967434e-001f,
	-7.401318e+000f,
	7.724021e+000f,
	-2.345723e-002f,
	5.979653e-001f,
	2.436346e-002f,
	2.658970e+000f,
	6.014891e-001f,
	-2.310933e+000f,
	-1.290290e+000f,
	-1.301909e+000f,
	2.557806e+000f,
	-3.744449e-002f,
	8.982861e-002f,
	1.090613e-001f,
	-4.398363e-001f,
	1.184329e-001f,
	-1.124730e+000f,
	-9.921830e-002f,
	1.366902e+000f,
	-4.172489e+000f,
	-5.078016e-002f,
	1.393597e+000f,
	-9.323843e-002f,
	6.452721e+000f,
	1.435913e+000f,
	-8.468477e-001f,
	-2.744819e-001f,
	-4.347200e-001f,
	6.713362e+000f,
	6.127133e-001f,
	-1.685634e+000f,
	7.360941e-001f,
	-4.535502e+000f,
	-2.920866e-002f,
	-1.165242e+000f,
	-2.008697e-001f,
	1.438778e+000f,
	-1.008936e+001f,
	-2.214771e+000f,
	2.102909e+000f,
	-1.763085e-001f,
	2.859075e+000f,
	1.093470e+000f,
	-1.074614e+000f,
	-2.066374e-001f,
	1.131891e+000f,
	1.630063e+001f,
	-4.801441e+000f,
	-1.112590e+000f,
	3.595785e-001f,
	1.122227e+000f,
	5.794610e-001f,
	// albedo 1, turbidity 9
	-1.521515e+000f,
	-6.835604e-001f,
	-5.571044e+000f,
	6.028774e+000f,
	-4.253715e-002f,
	6.875746e-001f,
	-5.279456e-006f,
	2.180150e+000f,
	8.487705e-001f,
	-2.240415e+000f,
	-1.171166e+000f,
	-7.182771e+000f,
	8.417068e+000f,
	-1.932866e-002f,
	1.101887e-001f,
	-1.098862e-002f,
	6.242195e-001f,
	-2.393875e-001f,
	-2.712354e+000f,
	-1.198830e+000f,
	3.180200e+000f,
	-6.768130e+000f,
	-2.563386e-003f,
	7.984607e-001f,
	2.764376e-001f,
	4.695358e+000f,
	1.557045e+000f,
	-3.655172e-002f,
	-2.142321e-002f,
	-9.138120e-001f,
	7.932786e+000f,
	3.516542e-001f,
	-7.994343e-001f,
	1.786761e-001f,
	-4.208399e+000f,
	1.820576e-002f,
	-1.368610e+000f,
	-2.656212e-001f,
	1.249397e+000f,
	-8.317818e+000f,
	-8.962772e-001f,
	1.423249e+000f,
	1.478381e-001f,
	2.191660e+000f,
	1.007748e+000f,
	-1.041753e+000f,
	-2.453366e-001f,
	1.061102e+000f,
	1.130172e+001f,
	-4.739312e+000f,
	-9.223334e-001f,
	2.982776e-001f,
	6.162931e-001f,
	6.080302e-001f,
	// albedo 1, turbidity 10
	-1.989159e+000f,
	-1.095160e+000f,
	-2.915550e+000f,
	3.275339e+000f,
	-5.735765e-002f,
	5.742174e-001f,
	-7.683288e-006f,
	1.763400e+000f,
	9.001342e-001f,
	-2.070020e+000f,
	-1.086338e+000f,
	-1.095898e+001f,
	1.206960e+001f,
	3.780123e-002f,
	-1.774699e-002f,
	-5.881348e-004f,
	1.333819e+000f,
	-2.605423e-001f,
	-5.249653e+000f,
	-2.383040e+000f,
	6.160406e+000f,
	-9.097138e+000f,
	-1.955319e-001f,
	1.651785e+000f,
	6.016463e-004f,
	3.021824e+000f,
	1.493574e+000f,
	4.685432e-001f,
	-2.358662e-001f,
	-2.666433e+000f,
	9.685763e+000f,
	5.804928e-001f,
	-1.521875e+000f,
	5.668989e-001f,
	-1.548136e+000f,
	1.688642e-002f,
	-1.296891e+000f,
	-3.449031e-001f,
	1.928548e+000f,
	-1.167560e+001f,
	-1.627615e+000f,
	1.355603e+000f,
	-1.929074e-001f,
	-6.568952e-001f,
	1.009774e+000f,
	-1.067288e+000f,
	-2.410392e-001f,
	7.147961e-001f,
	1.783840e+001f,
	-4.374399e+000f,
	-6.588777e-001f,
	3.329831e-001f,
	1.012066e+000f,
	6.118645e-001f,
};

float datasetXYZRad2[] = 
{
	// albedo 0, turbidity 1
	1.632341e+000f,
	1.395230e+000f,
	1.375634e+000f,
	1.238193e+001f,
	5.921102e+000f,
	7.766508e+000f,
	// albedo 0, turbidity 2
	1.597115e+000f,
	1.554617e+000f,
	3.932382e-001f,
	1.505284e+001f,
	5.725234e+000f,
	8.158155e+000f,
	// albedo 0, turbidity 3
	1.522034e+000f,
	1.844545e+000f,
	-1.322862e+000f,
	1.918382e+001f,
	5.440769e+000f,
	8.837119e+000f,
	// albedo 0, turbidity 4
	1.403048e+000f,
	2.290852e+000f,
	-4.013792e+000f,
	2.485100e+001f,
	5.521888e+000f,
	9.845547e+000f,
	// albedo 0, turbidity 5
	1.286364e+000f,
	2.774498e+000f,
	-6.648221e+000f,
	2.964151e+001f,
	5.923777e+000f,
	1.097075e+001f,
	// albedo 0, turbidity 6
	1.213544e+000f,
	3.040195e+000f,
	-8.092676e+000f,
	3.186082e+001f,
	6.789782e+000f,
	1.158899e+001f,
	// albedo 0, turbidity 7
	1.122622e+000f,
	3.347465e+000f,
	-9.649016e+000f,
	3.343824e+001f,
	9.347715e+000f,
	1.231374e+001f,
	// albedo 0, turbidity 8
	1.007356e+000f,
	3.543858e+000f,
	-1.053520e+001f,
	3.239842e+001f,
	1.483962e+001f,
	1.331718e+001f,
	// albedo 0, turbidity 9
	8.956642e-001f,
	3.278700e+000f,
	-9.254933e+000f,
	2.557923e+001f,
	2.489677e+001f,
	1.476166e+001f,
	// albedo 0, turbidity 10
	7.985143e-001f,
	2.340404e+000f,
	-4.928274e+000f,
	1.141787e+001f,
	3.961501e+001f,
	1.682448e+001f,
	// albedo 1, turbidity 1
	1.745162e+000f,
	1.639467e+000f,
	1.342721e+000f,
	1.166033e+001f,
	1.490124e+001f,
	1.774031e+001f,
	// albedo 1, turbidity 2
	1.708439e+000f,
	1.819144e+000f,
	2.834399e-001f,
	1.448066e+001f,
	1.459214e+001f,
	1.858679e+001f,
	// albedo 1, turbidity 3
	1.631720e+000f,
	2.094799e+000f,
	-1.378825e+000f,
	1.843198e+001f,
	1.463173e+001f,
	1.962881e+001f,
	// albedo 1, turbidity 4
	1.516536e+000f,
	2.438729e+000f,
	-3.624121e+000f,
	2.298621e+001f,
	1.599782e+001f,
	2.070027e+001f,
	// albedo 1, turbidity 5
	1.405863e+000f,
	2.785191e+000f,
	-5.705236e+000f,
	2.645121e+001f,
	1.768330e+001f,
	2.191903e+001f,
	// albedo 1, turbidity 6
	1.344052e+000f,
	2.951807e+000f,
	-6.683851e+000f,
	2.744271e+001f,
	1.985706e+001f,
	2.229452e+001f,
	// albedo 1, turbidity 7
	1.245827e+000f,
	3.182923e+000f,
	-7.822960e+000f,
	2.791395e+001f,
	2.327254e+001f,
	2.315910e+001f,
	// albedo 1, turbidity 8
	1.132305e+000f,
	3.202593e+000f,
	-8.008429e+000f,
	2.521093e+001f,
	3.000014e+001f,
	2.405306e+001f,
	// albedo 1, turbidity 9
	1.020330e+000f,
	2.820556e+000f,
	-6.238704e+000f,
	1.709276e+001f,
	4.077916e+001f,
	2.509949e+001f,
	// albedo 1, turbidity 10
	9.031570e-001f,
	1.863917e+000f,
	-1.955738e+000f,
	3.032665e+000f,
	5.434290e+001f,
	2.641780e+001f,
};

float datasetXYZ3[] = 
{
	// albedo 0, turbidity 1
	-1.310023e+000f,
	-4.407658e-001f,
	-3.640340e+001f,
	3.683292e+001f,
	-8.124762e-003f,
	5.297961e-001f,
	1.188633e-002f,
	3.138320e+000f,
	5.134778e-001f,
	-1.424100e+000f,
	-5.501606e-001f,
	-1.753510e+001f,
	1.822769e+001f,
	-1.539272e-002f,
	6.366826e-001f,
	2.661996e-003f,
	2.659915e+000f,
	4.071138e-001f,
	-1.103436e+000f,
	-1.884105e-001f,
	6.425322e+000f,
	-6.910579e+000f,
	-2.019861e-002f,
	3.553271e-001f,
	-1.589061e-002f,
	5.345985e+000f,
	8.790218e-001f,
	-1.186200e+000f,
	-4.307514e-001f,
	-3.957947e+000f,
	5.979352e+000f,
	-5.348869e-002f,
	1.736117e+000f,
	3.491346e-002f,
	-2.692261e+000f,
	5.610506e-001f,
	-1.006038e+000f,
	-1.305995e-001f,
	4.473513e+000f,
	-3.806719e+000f,
	1.419407e-001f,
	-2.148238e-002f,
	-5.081185e-002f,
	3.735362e+000f,
	5.358280e-001f,
	-1.078507e+000f,
	-1.633754e-001f,
	-3.812368e+000f,
	4.381700e+000f,
	2.988122e-002f,
	1.754224e+000f,
	1.472376e-001f,
	3.722798e+000f,
	4.999157e-001f,
	// albedo 0, turbidity 2
	-1.333582e+000f,
	-4.649908e-001f,
	-3.359528e+001f,
	3.404375e+001f,
	-9.384242e-003f,
	5.587511e-001f,
	5.726310e-003f,
	3.073145e+000f,
	5.425529e-001f,
	-1.562624e+000f,
	-7.107068e-001f,
	-1.478170e+001f,
	1.559839e+001f,
	-1.462375e-002f,
	5.050133e-001f,
	2.516017e-002f,
	1.604696e+000f,
	2.902403e-001f,
	-8.930158e-001f,
	4.068077e-002f,
	1.373481e+000f,
	-2.342752e+000f,
	-2.098058e-002f,
	6.248686e-001f,
	-5.258363e-002f,
	7.058214e+000f,
	1.150373e+000f,
	-1.262823e+000f,
	-4.818353e-001f,
	8.892610e-004f,
	1.923120e+000f,
	-4.979718e-002f,
	1.040693e+000f,
	1.558103e-001f,
	-2.852480e+000f,
	2.420691e-001f,
	-9.968383e-001f,
	-1.200648e-001f,
	1.324342e+000f,
	-9.430889e-001f,
	1.931098e-001f,
	4.436916e-001f,
	-7.320456e-002f,
	4.215931e+000f,
	7.898019e-001f,
	-1.078185e+000f,
	-1.718192e-001f,
	-1.720191e+000f,
	2.358918e+000f,
	2.765637e-002f,
	1.260245e+000f,
	2.021941e-001f,
	3.395483e+000f,
	5.173628e-001f,
	// albedo 0, turbidity 3
	-1.353023e+000f,
	-4.813523e-001f,
	-3.104920e+001f,
	3.140156e+001f,
	-9.510741e-003f,
	5.542030e-001f,
	8.135471e-003f,
	3.136646e+000f,
	5.215989e-001f,
	-1.624704e+000f,
	-7.990201e-001f,
	-2.167125e+001f,
	2.246341e+001f,
	-1.163533e-002f,
	5.415746e-001f,
	2.618378e-002f,
	1.139214e+000f,
	3.444357e-001f,
	-7.983610e-001f,
	1.417476e-001f,
	9.914841e+000f,
	-1.081503e+001f,
	-1.218845e-002f,
	3.411392e-001f,
	-6.137698e-002f,
	7.445848e+000f,
	1.180080e+000f,
	-1.266679e+000f,
	-4.288977e-001f,
	-5.818701e+000f,
	6.986437e+000f,
	-8.180711e-002f,
	1.397403e+000f,
	2.016916e-001f,
	-1.275731e+000f,
	2.592773e-001f,
	-1.009707e+000f,
	-1.537754e-001f,
	3.496378e+000f,
	-3.013726e+000f,
	2.421150e-001f,
	-2.831925e-001f,
	3.003395e-002f,
	3.702862e+000f,
	7.746320e-001f,
	-1.075646e+000f,
	-1.768747e-001f,
	-1.347762e+000f,
	1.989004e+000f,
	1.375836e-002f,
	1.764810e+000f,
	1.330018e-001f,
	3.230864e+000f,
	6.626210e-001f,
	// albedo 0, turbidity 4
	-1.375269e+000f,
	-5.103569e-001f,
	-3.442661e+001f,
	3.478703e+001f,
	-8.460009e-003f,
	5.408643e-001f,
	4.813323e-003f,
	3.016078e+000f,
	5.062069e-001f,
	-1.821679e+000f,
	-9.766461e-001f,
	-1.926488e+001f,
	1.997912e+001f,
	-9.822567e-003f,
	3.649556e-001f,
	4.316092e-002f,
	8.930190e-001f,
	4.166527e-001f,
	-6.633542e-001f,
	1.997841e-001f,
	2.395592e+000f,
	-3.117175e+000f,
	-1.080884e-002f,
	8.983814e-001f,
	-1.375825e-001f,
	6.673463e+000f,
	1.115663e+000f,
	-1.303240e+000f,
	-3.612712e-001f,
	8.292959e-002f,
	3.381364e-001f,
	-6.078648e-002f,
	3.229247e-001f,
	3.680987e-001f,
	7.046755e-001f,
	3.144924e-001f,
	-9.952598e-001f,
	-2.039076e-001f,
	4.026851e-001f,
	2.686684e-001f,
	1.640712e-001f,
	5.186341e-001f,
	-1.205520e-002f,
	2.659613e+000f,
	8.030394e-001f,
	-1.098579e+000f,
	-2.151992e-001f,
	6.558198e-001f,
	-7.436900e-004f,
	-1.421817e-003f,
	1.073701e+000f,
	1.886875e-001f,
	2.536857e+000f,
	6.673923e-001f,
	// albedo 0, turbidity 5
	-1.457986e+000f,
	-5.906842e-001f,
	-3.812464e+001f,
	3.838539e+001f,
	-6.024357e-003f,
	4.741484e-001f,
	1.209223e-002f,
	2.818432e+000f,
	5.012433e-001f,
	-1.835728e+000f,
	-1.003405e+000f,
	-6.848129e+000f,
	7.601943e+000f,
	-1.277375e-002f,
	4.785598e-001f,
	3.366853e-002f,
	1.097701e+000f,
	4.636635e-001f,
	-8.491348e-001f,
	9.466365e-003f,
	-2.685226e+000f,
	2.004060e+000f,
	-1.168708e-002f,
	6.752316e-001f,
	-1.543371e-001f,
	5.674759e+000f,
	1.039534e+000f,
	-1.083379e+000f,
	-1.506790e-001f,
	7.328236e-001f,
	-5.095568e-001f,
	-8.609153e-002f,
	4.448820e-001f,
	4.174662e-001f,
	1.481556e+000f,
	3.942551e-001f,
	-1.117089e+000f,
	-3.337605e-001f,
	2.502281e-001f,
	4.036323e-001f,
	2.673899e-001f,
	2.829817e-001f,
	2.242450e-002f,
	2.043207e+000f,
	7.706902e-001f,
	-1.071648e+000f,
	-2.126200e-001f,
	6.069466e-001f,
	-1.456290e-003f,
	-5.515960e-001f,
	1.046755e+000f,
	1.985021e-001f,
	2.290245e+000f,
	6.876058e-001f,
	// albedo 0, turbidity 6
	-1.483903e+000f,
	-6.309647e-001f,
	-4.380213e+001f,
	4.410537e+001f,
	-5.712161e-003f,
	5.195992e-001f,
	2.028428e-003f,
	2.687114e+000f,
	5.098321e-001f,
	-2.053976e+000f,
	-1.141473e+000f,
	5.109183e-001f,
	8.060391e-002f,
	-1.033983e-002f,
	4.066532e-001f,
	4.869627e-002f,
	1.161722e+000f,
	4.039525e-001f,
	-6.348185e-001f,
	7.651292e-002f,
	-1.031327e+001f,
	1.007598e+001f,
	-2.083688e-002f,
	7.359516e-001f,
	-2.029459e-001f,
	5.013257e+000f,
	1.077649e+000f,
	-1.228630e+000f,
	-1.650496e-001f,
	4.077157e-002f,
	-7.189167e-001f,
	-5.092220e-002f,
	2.959814e-001f,
	5.111496e-001f,
	2.540433e+000f,
	3.615330e-001f,
	-1.041883e+000f,
	-3.278413e-001f,
	-6.691911e-002f,
	1.307364e+000f,
	2.166663e-001f,
	3.000595e-001f,
	-3.157136e-003f,
	1.389208e+000f,
	7.999026e-001f,
	-1.103556e+000f,
	-2.443602e-001f,
	4.705347e-001f,
	-9.296482e-004f,
	-5.309920e-001f,
	9.654511e-001f,
	2.142587e-001f,
	2.244723e+000f,
	6.839976e-001f,
	// albedo 0, turbidity 7
	-1.555684e+000f,
	-6.962113e-001f,
	-4.647983e+001f,
	4.674270e+001f,
	-5.034895e-003f,
	4.755090e-001f,
	-9.502561e-007f,
	2.626569e+000f,
	5.056194e-001f,
	-1.998288e+000f,
	-1.124720e+000f,
	-1.629586e+000f,
	2.187993e+000f,
	-8.284384e-003f,
	3.845258e-001f,
	5.726240e-002f,
	1.185644e+000f,
	4.255812e-001f,
	-1.032570e+000f,
	-2.513850e-001f,
	-3.721112e+000f,
	3.506967e+000f,
	-2.186561e-002f,
	9.436049e-001f,
	-2.451412e-001f,
	4.725724e+000f,
	1.039256e+000f,
	-8.597532e-001f,
	9.073332e-002f,
	-2.553741e+000f,
	1.993237e+000f,
	-4.390891e-002f,
	-2.046928e-001f,
	5.515623e-001f,
	1.909127e+000f,
	3.948212e-001f,
	-1.210482e+000f,
	-4.477622e-001f,
	-2.267805e-001f,
	1.219488e+000f,
	1.336186e-001f,
	6.866897e-001f,
	2.808997e-002f,
	1.600403e+000f,
	7.816409e-001f,
	-1.078168e+000f,
	-2.699261e-001f,
	2.537282e-001f,
	3.820684e-001f,
	-4.425103e-001f,
	5.298235e-001f,
	2.185217e-001f,
	1.728679e+000f,
	6.882743e-001f,
	// albedo 0, turbidity 8
	-1.697968e+000f,
	-8.391488e-001f,
	-5.790105e+001f,
	5.814120e+001f,
	-3.404760e-003f,
	4.265140e-001f,
	-1.796301e-006f,
	2.368442e+000f,
	5.324429e-001f,
	-2.141552e+000f,
	-1.172230e+000f,
	1.677872e+001f,
	-1.641470e+001f,
	-5.732425e-003f,
	2.002199e-001f,
	6.841834e-002f,
	1.485338e+000f,
	3.215763e-001f,
	-1.442946e+000f,
	-7.264245e-001f,
	-9.503706e+000f,
	9.650462e+000f,
	-2.120995e-002f,
	1.419263e+000f,
	-2.893098e-001f,
	3.860731e+000f,
	1.120857e+000f,
	-5.696752e-001f,
	3.411279e-001f,
	-2.931035e-001f,
	-6.512552e-001f,
	-1.068437e-001f,
	-1.085661e+000f,
	6.107549e-001f,
	1.459503e+000f,
	3.210336e-001f,
	-1.313839e+000f,
	-5.921371e-001f,
	-2.332222e-001f,
	1.648196e+000f,
	2.492787e-001f,
	1.381033e+000f,
	-1.993392e-002f,
	9.812560e-001f,
	8.316329e-001f,
	-1.087464e+000f,
	-3.195534e-001f,
	2.902095e-001f,
	3.383709e-001f,
	-8.798482e-001f,
	1.494668e-002f,
	2.529703e-001f,
	1.452644e+000f,
	6.693870e-001f,
	// albedo 0, turbidity 9
	-2.068582e+000f,
	-1.118605e+000f,
	-5.081598e+001f,
	5.097486e+001f,
	-3.280669e-003f,
	4.067371e-001f,
	-2.544951e-006f,
	2.179497e+000f,
	5.778017e-001f,
	-1.744693e+000f,
	-8.537207e-001f,
	2.234361e+001f,
	-2.208318e+001f,
	-5.932616e-003f,
	1.035049e-001f,
	5.742772e-002f,
	1.977880e+000f,
	2.124846e-001f,
	-3.287515e+000f,
	-2.140268e+000f,
	-1.249566e+001f,
	1.240091e+001f,
	-2.409349e-002f,
	1.397821e+000f,
	-2.371627e-001f,
	2.771192e+000f,
	1.170496e+000f,
	5.502311e-001f,
	1.046630e+000f,
	2.193517e+000f,
	-2.220400e+000f,
	-1.064394e-001f,
	-1.017926e+000f,
	4.795457e-001f,
	1.030644e+000f,
	3.177516e-001f,
	-1.719734e+000f,
	-9.536198e-001f,
	-6.586821e-001f,
	1.386361e+000f,
	-2.513065e-002f,
	1.187011e+000f,
	6.542539e-002f,
	5.296055e-001f,
	8.082660e-001f,
	-1.005700e+000f,
	-3.028096e-001f,
	4.470957e-002f,
	1.007760e+000f,
	-8.119016e-001f,
	3.153338e-002f,
	2.311321e-001f,
	1.182208e+000f,
	6.824758e-001f,
	// albedo 0, turbidity 10
	-2.728867e+000f,
	-1.580388e+000f,
	-3.079627e+001f,
	3.092586e+001f,
	-4.197673e-003f,
	3.154759e-001f,
	-3.897675e-006f,
	1.920567e+000f,
	6.664791e-001f,
	-1.322495e+000f,
	-7.249275e-001f,
	1.477660e+001f,
	-1.468154e+001f,
	-9.044857e-003f,
	5.624314e-002f,
	6.498392e-002f,
	2.047389e+000f,
	6.367540e-002f,
	-6.102376e+000f,
	-3.473018e+000f,
	-9.926071e+000f,
	9.637797e+000f,
	-1.097909e-002f,
	1.103498e+000f,
	-2.424521e-001f,
	2.520748e+000f,
	1.240260e+000f,
	1.351796e+000f,
	1.018588e+000f,
	2.009081e+000f,
	-1.333394e+000f,
	-1.979125e-001f,
	-3.318292e-001f,
	4.476624e-001f,
	9.095235e-001f,
	2.955611e-001f,
	-1.774467e+000f,
	-1.079880e+000f,
	-8.084680e-002f,
	2.577697e-001f,
	-1.149295e-001f,
	4.975303e-001f,
	2.931611e-003f,
	-3.803171e-001f,
	8.002794e-001f,
	-9.898401e-001f,
	-2.542513e-001f,
	-7.530911e-002f,
	1.870355e+000f,
	-1.521918e+000f,
	2.405164e-001f,
	2.964615e-001f,
	1.334800e+000f,
	6.789053e-001f,
	// albedo 1, turbidity 1
	-1.279730e+000f,
	-4.290674e-001f,
	-4.277972e+001f,
	4.343305e+001f,
	-6.541826e-003f,
	4.945086e-001f,
	1.425338e-002f,
	2.685244e+000f,
	5.011313e-001f,
	-1.449506e+000f,
	-5.766374e-001f,
	-1.688496e+001f,
	1.781118e+001f,
	-1.121649e-002f,
	3.545020e-001f,
	2.287338e-002f,
	1.904281e+000f,
	4.936998e-001f,
	-1.021980e+000f,
	-1.897574e-001f,
	2.482462e+000f,
	-2.941725e+000f,
	-1.570448e-002f,
	7.532578e-001f,
	-4.256800e-002f,
	5.239660e+000f,
	4.983116e-001f,
	-1.162608e+000f,
	-3.428049e-001f,
	3.974358e+000f,
	-1.527935e+000f,
	-3.919201e-002f,
	8.758593e-001f,
	7.291363e-002f,
	-3.455257e+000f,
	8.007426e-001f,
	-9.929985e-001f,
	-8.712006e-002f,
	-7.397313e-001f,
	1.348372e+000f,
	9.511685e-002f,
	3.233584e-001f,
	-7.549148e-002f,
	5.806452e+000f,
	4.990042e-001f,
	-1.084996e+000f,
	-1.739767e-001f,
	1.580475e-001f,
	9.088180e-001f,
	6.871433e-002f,
	5.933079e-001f,
	1.188921e-001f,
	3.074079e+000f,
	4.999327e-001f,
	// albedo 1, turbidity 2
	-1.317009e+000f,
	-4.661946e-001f,
	-4.255347e+001f,
	4.312782e+001f,
	-5.727235e-003f,
	4.285447e-001f,
	2.189854e-002f,
	2.608310e+000f,
	5.190700e-001f,
	-1.469236e+000f,
	-6.282139e-001f,
	-1.241404e+001f,
	1.348765e+001f,
	-1.204770e-002f,
	5.070285e-001f,
	-7.280216e-004f,
	1.491533e+000f,
	3.635064e-001f,
	-9.713808e-001f,
	-8.138038e-002f,
	3.709854e-001f,
	-1.041174e+000f,
	-1.814075e-002f,
	5.060860e-001f,
	-2.053756e-002f,
	6.161431e+000f,
	1.093736e+000f,
	-1.159057e+000f,
	-3.698074e-001f,
	2.711209e+000f,
	-6.006479e-001f,
	-4.896926e-002f,
	9.273957e-001f,
	1.137712e-001f,
	-3.496828e+000f,
	2.867109e-001f,
	-1.011601e+000f,
	-8.201890e-002f,
	2.105725e-001f,
	4.597520e-001f,
	1.478925e-001f,
	2.138940e-001f,
	-5.660670e-002f,
	6.057755e+000f,
	7.859121e-001f,
	-1.078020e+000f,
	-1.811580e-001f,
	1.646622e-001f,
	8.348426e-001f,
	1.149064e-001f,
	4.985738e-001f,
	1.376605e-001f,
	2.746607e+000f,
	4.999626e-001f,
	// albedo 1, turbidity 3
	-1.325672e+000f,
	-4.769313e-001f,
	-4.111215e+001f,
	4.168293e+001f,
	-6.274997e-003f,
	4.649469e-001f,
	1.119411e-002f,
	2.631267e+000f,
	5.234546e-001f,
	-1.619391e+000f,
	-8.000253e-001f,
	-1.534098e+001f,
	1.632706e+001f,
	-1.012023e-002f,
	4.242255e-001f,
	2.931597e-002f,
	8.925807e-001f,
	3.314765e-001f,
	-7.356979e-001f,
	1.368406e-001f,
	2.972579e+000f,
	-3.535359e+000f,
	-1.318948e-002f,
	4.607620e-001f,
	-7.182778e-002f,
	6.254100e+000f,
	1.236299e+000f,
	-1.316217e+000f,
	-4.194427e-001f,
	3.489902e-002f,
	1.289849e+000f,
	-4.755960e-002f,
	1.138222e+000f,
	1.975992e-001f,
	-8.991542e-001f,
	2.290572e-001f,
	-9.502188e-001f,
	-1.172703e-001f,
	1.405202e+000f,
	-3.061919e-001f,
	1.058772e-001f,
	-3.760592e-001f,
	-1.983179e-002f,
	3.562353e+000f,
	7.895959e-001f,
	-1.100117e+000f,
	-1.900567e-001f,
	4.925030e-001f,
	5.250225e-001f,
	1.576804e-001f,
	1.042701e+000f,
	7.330743e-002f,
	2.796064e+000f,
	6.749783e-001f,
	// albedo 1, turbidity 4
	-1.354183e+000f,
	-5.130625e-001f,
	-4.219268e+001f,
	4.271772e+001f,
	-5.365373e-003f,
	4.136743e-001f,
	1.235172e-002f,
	2.520122e+000f,
	5.187269e-001f,
	-1.741434e+000f,
	-9.589761e-001f,
	-8.230339e+000f,
	9.296799e+000f,
	-9.600162e-003f,
	4.994969e-001f,
	2.955452e-002f,
	3.667099e-001f,
	3.526999e-001f,
	-6.917347e-001f,
	2.154887e-001f,
	-8.760264e-001f,
	2.334121e-001f,
	-1.909621e-002f,
	4.748033e-001f,
	-1.138514e-001f,
	6.515360e+000f,
	1.225097e+000f,
	-1.293189e+000f,
	-4.218700e-001f,
	1.620952e+000f,
	-7.858597e-001f,
	-3.769410e-002f,
	6.636786e-001f,
	3.364945e-001f,
	-5.341017e-001f,
	2.128347e-001f,
	-9.735521e-001f,
	-1.325495e-001f,
	1.007517e+000f,
	2.598258e-001f,
	6.762169e-002f,
	1.421018e-003f,
	-6.915987e-002f,
	3.185897e+000f,
	8.641956e-001f,
	-1.094800e+000f,
	-1.962062e-001f,
	5.755591e-001f,
	2.906259e-001f,
	2.625748e-001f,
	7.644049e-001f,
	1.347492e-001f,
	2.677126e+000f,
	6.465460e-001f,
	// albedo 1, turbidity 5
	-1.393063e+000f,
	-5.578338e-001f,
	-4.185249e+001f,
	4.233504e+001f,
	-5.435640e-003f,
	4.743765e-001f,
	7.422477e-003f,
	2.442801e+000f,
	5.211707e-001f,
	-1.939487e+000f,
	-1.128509e+000f,
	-8.974257e+000f,
	9.978383e+000f,
	-7.965597e-003f,
	2.948830e-001f,
	4.436763e-002f,
	2.839868e-001f,
	3.440424e-001f,
	-6.011562e-001f,
	2.354877e-001f,
	-3.079820e+000f,
	2.585094e+000f,
	-2.002701e-002f,
	7.793909e-001f,
	-1.598414e-001f,
	5.834678e+000f,
	1.202856e+000f,
	-1.315676e+000f,
	-3.903446e-001f,
	1.701900e+000f,
	-1.304609e+000f,
	-1.045121e-002f,
	2.747707e-001f,
	4.143967e-001f,
	3.197102e-001f,
	2.637580e-001f,
	-9.618628e-001f,
	-1.625841e-001f,
	1.187138e+000f,
	1.497802e-001f,
	-5.590954e-006f,
	3.178475e-002f,
	-4.153145e-002f,
	2.496096e+000f,
	8.195082e-001f,
	-1.111554e+000f,
	-2.365546e-001f,
	7.831875e-001f,
	2.018684e-001f,
	2.074369e-001f,
	7.395978e-001f,
	1.225730e-001f,
	1.876478e+000f,
	6.821167e-001f,
	// albedo 1, turbidity 6
	-1.427879e+000f,
	-5.994879e-001f,
	-3.531016e+001f,
	3.581581e+001f,
	-6.431497e-003f,
	4.554192e-001f,
	7.348731e-004f,
	2.334619e+000f,
	5.233377e-001f,
	-1.998177e+000f,
	-1.206633e+000f,
	-2.146510e+001f,
	2.242237e+001f,
	-5.857596e-003f,
	2.755663e-001f,
	6.384795e-002f,
	1.358244e-001f,
	3.328437e-001f,
	-6.440630e-001f,
	2.058571e-001f,
	2.155499e+000f,
	-2.587968e+000f,
	-1.840023e-002f,
	8.826555e-001f,
	-2.222452e-001f,
	5.847073e+000f,
	1.228387e+000f,
	-1.229071e+000f,
	-3.360441e-001f,
	-3.429599e-001f,
	6.179469e-001f,
	2.029610e-003f,
	8.899319e-002f,
	5.041624e-001f,
	1.882964e-001f,
	2.252040e-001f,
	-1.022905e+000f,
	-2.101621e-001f,
	1.915689e+000f,
	-6.498794e-001f,
	-3.463651e-002f,
	8.954605e-002f,
	-6.797854e-002f,
	2.417705e+000f,
	8.568618e-001f,
	-1.082538e+000f,
	-2.007723e-001f,
	4.731009e-001f,
	4.077267e-001f,
	1.324289e-001f,
	6.514880e-001f,
	1.702912e-001f,
	2.309383e+000f,
	6.600895e-001f,
	// albedo 1, turbidity 7
	-1.472139e+000f,
	-6.499815e-001f,
	-3.428465e+001f,
	3.469659e+001f,
	-5.747023e-003f,
	4.174167e-001f,
	1.688597e-003f,
	2.323046e+000f,
	5.395191e-001f,
	-2.161176e+000f,
	-1.353089e+000f,
	-2.226827e+001f,
	2.329138e+001f,
	-5.583808e-003f,
	2.364793e-001f,
	6.096656e-002f,
	1.944666e-003f,
	2.861624e-001f,
	-6.593044e-001f,
	1.393558e-001f,
	4.698373e+000f,
	-5.193883e+000f,
	-1.998390e-002f,
	1.095635e+000f,
	-2.391254e-001f,
	5.598103e+000f,
	1.236193e+000f,
	-1.195717e+000f,
	-2.972715e-001f,
	4.648953e-002f,
	3.024588e-001f,
	5.003313e-003f,
	-3.754741e-001f,
	5.247265e-001f,
	-1.381312e-001f,
	2.493896e-001f,
	-1.020139e+000f,
	-2.253524e-001f,
	3.548437e-001f,
	7.030485e-001f,
	-2.107076e-002f,
	4.581395e-001f,
	-3.243757e-002f,
	2.453259e+000f,
	8.323623e-001f,
	-1.098770e+000f,
	-2.435780e-001f,
	8.761614e-001f,
	1.941613e-001f,
	-1.990692e-001f,
	3.761139e-001f,
	1.657412e-001f,
	1.590503e+000f,
	6.741417e-001f,
	// albedo 1, turbidity 8
	-1.648007e+000f,
	-8.205121e-001f,
	-4.435106e+001f,
	4.479801e+001f,
	-4.181353e-003f,
	3.854830e-001f,
	-1.842385e-006f,
	2.000281e+000f,
	5.518363e-001f,
	-2.140986e+000f,
	-1.282239e+000f,
	-3.979213e+000f,
	4.672459e+000f,
	-5.008582e-003f,
	2.421920e-001f,
	6.253602e-002f,
	6.612713e-001f,
	2.555851e-001f,
	-1.300502e+000f,
	-5.137898e-001f,
	5.179821e-001f,
	-4.032341e-001f,
	-2.066785e-002f,
	1.087929e+000f,
	-2.615309e-001f,
	4.225887e+000f,
	1.229237e+000f,
	-6.963340e-001f,
	9.241060e-002f,
	6.936356e-002f,
	-3.588571e-001f,
	-5.461843e-002f,
	-5.616643e-001f,
	5.484166e-001f,
	-4.776267e-002f,
	2.414935e-001f,
	-1.233179e+000f,
	-4.325498e-001f,
	6.479813e-001f,
	8.368356e-001f,
	2.458875e-001f,
	6.464752e-001f,
	-2.897097e-002f,
	1.561773e+000f,
	8.518598e-001f,
	-1.051023e+000f,
	-2.533690e-001f,
	1.004294e+000f,
	3.028083e-001f,
	-1.520108e+000f,
	1.607013e-001f,
	1.619975e-001f,
	1.131094e+000f,
	6.706655e-001f,
	// albedo 1, turbidity 9
	-1.948249e+000f,
	-1.097383e+000f,
	-4.453697e+001f,
	4.494902e+001f,
	-3.579939e-003f,
	3.491605e-001f,
	-2.500253e-006f,
	1.740442e+000f,
	6.188022e-001f,
	-2.154253e+000f,
	-1.209559e+000f,
	4.144894e+000f,
	-3.562411e+000f,
	-5.638843e-003f,
	1.067169e-001f,
	7.594858e-002f,
	1.005280e+000f,
	1.072543e-001f,
	-2.513259e+000f,
	-1.507208e+000f,
	-1.602979e+000f,
	1.404154e+000f,
	-5.560750e-003f,
	1.240490e+000f,
	-2.852117e-001f,
	3.485252e+000f,
	1.349321e+000f,
	-7.832214e-002f,
	3.655626e-001f,
	3.856288e-001f,
	6.867894e-001f,
	-1.609523e-001f,
	-6.704306e-001f,
	5.357301e-001f,
	-6.457935e-001f,
	1.479503e-001f,
	-1.354784e+000f,
	-5.454375e-001f,
	8.797469e-001f,
	-1.466514e+000f,
	7.134420e-001f,
	5.934903e-001f,
	-2.911178e-002f,
	8.643737e-001f,
	9.030724e-001f,
	-1.048324e+000f,
	-2.738736e-001f,
	8.783074e-001f,
	3.246188e+000f,
	-4.435369e+000f,
	1.251791e-001f,
	1.783486e-001f,
	1.064657e+000f,
	6.522878e-001f,
	// albedo 1, turbidity 10
	-2.770408e+000f,
	-1.618911e+000f,
	-2.504031e+001f,
	2.531674e+001f,
	-4.239279e-003f,
	3.241013e-001f,
	-3.764484e-006f,
	1.586843e+000f,
	7.035906e-001f,
	-1.913500e+000f,
	-1.144014e+000f,
	-1.080587e+001f,
	1.153677e+001f,
	-1.003197e-002f,
	1.577515e-001f,
	5.217789e-002f,
	1.225278e+000f,
	5.172771e-003f,
	-5.293208e+000f,
	-2.876463e+000f,
	2.087053e+000f,
	-3.201552e+000f,
	3.892964e-003f,
	5.323930e-001f,
	-2.034512e-001f,
	2.617760e+000f,
	1.273597e+000f,
	9.060340e-001f,
	3.773409e-001f,
	-6.399945e-001f,
	3.213979e+000f,
	-9.112172e-002f,
	6.494055e-001f,
	3.953280e-001f,
	5.047796e-001f,
	2.998695e-001f,
	-1.482179e+000f,
	-6.778310e-001f,
	1.161775e+000f,
	-3.004872e+000f,
	4.774797e-001f,
	-4.969248e-001f,
	-3.512074e-003f,
	-1.307190e+000f,
	7.927378e-001f,
	-9.863181e-001f,
	-1.803364e-001f,
	5.810824e-001f,
	4.580570e+000f,
	-3.863454e+000f,
	5.328174e-001f,
	2.272821e-001f,
	1.771114e+000f,
	6.791814e-001f,
};

float datasetXYZRad3[] = 
{
	// albedo 0, turbidity 1
	1.168084e+000f,
	2.156455e+000f,
	-3.980314e+000f,
	1.989302e+001f,
	1.328335e+001f,
	1.435621e+001f,
	// albedo 0, turbidity 2
	1.135488e+000f,
	2.294701e+000f,
	-4.585886e+000f,
	2.090208e+001f,
	1.347840e+001f,
	1.467658e+001f,
	// albedo 0, turbidity 3
	1.107408e+000f,
	2.382765e+000f,
	-5.112357e+000f,
	2.147823e+001f,
	1.493128e+001f,
	1.460882e+001f,
	// albedo 0, turbidity 4
	1.054193e+000f,
	2.592891e+000f,
	-6.115000e+000f,
	2.268967e+001f,
	1.635672e+001f,
	1.518999e+001f,
	// albedo 0, turbidity 5
	1.006946e+000f,
	2.705420e+000f,
	-6.698930e+000f,
	2.291830e+001f,
	1.834324e+001f,
	1.570651e+001f,
	// albedo 0, turbidity 6
	9.794044e-001f,
	2.742440e+000f,
	-6.805283e+000f,
	2.225271e+001f,
	2.050797e+001f,
	1.563130e+001f,
	// albedo 0, turbidity 7
	9.413577e-001f,
	2.722009e+000f,
	-6.760707e+000f,
	2.098242e+001f,
	2.342588e+001f,
	1.605011e+001f,
	// albedo 0, turbidity 8
	8.917923e-001f,
	2.592780e+000f,
	-6.152635e+000f,
	1.774141e+001f,
	2.858324e+001f,
	1.657910e+001f,
	// albedo 0, turbidity 9
	8.288391e-001f,
	2.153434e+000f,
	-4.118327e+000f,
	1.078118e+001f,
	3.681710e+001f,
	1.738139e+001f,
	// albedo 0, turbidity 10
	7.623528e-001f,
	1.418187e+000f,
	-8.845235e-001f,
	7.590129e-001f,
	4.629859e+001f,
	1.921657e+001f,
	// albedo 1, turbidity 1
	1.352858e+000f,
	2.048862e+000f,
	-2.053393e+000f,
	1.405874e+001f,
	3.045344e+001f,
	3.044430e+001f,
	// albedo 1, turbidity 2
	1.330497e+000f,
	2.126497e+000f,
	-2.466296e+000f,
	1.467559e+001f,
	3.090738e+001f,
	3.069707e+001f,
	// albedo 1, turbidity 3
	1.286344e+000f,
	2.200436e+000f,
	-2.877228e+000f,
	1.492701e+001f,
	3.236288e+001f,
	3.077223e+001f,
	// albedo 1, turbidity 4
	1.234428e+000f,
	2.289628e+000f,
	-3.404699e+000f,
	1.499436e+001f,
	3.468390e+001f,
	3.084842e+001f,
	// albedo 1, turbidity 5
	1.178660e+000f,
	2.306071e+000f,
	-3.549159e+000f,
	1.411006e+001f,
	3.754188e+001f,
	3.079730e+001f,
	// albedo 1, turbidity 6
	1.151366e+000f,
	2.333005e+000f,
	-3.728627e+000f,
	1.363374e+001f,
	3.905894e+001f,
	3.092599e+001f,
	// albedo 1, turbidity 7
	1.101593e+000f,
	2.299422e+000f,
	-3.565787e+000f,
	1.196745e+001f,
	4.188472e+001f,
	3.102755e+001f,
	// albedo 1, turbidity 8
	1.038322e+000f,
	2.083539e+000f,
	-2.649585e+000f,
	8.037389e+000f,
	4.700869e+001f,
	3.065948e+001f,
	// albedo 1, turbidity 9
	9.596146e-001f,
	1.671470e+000f,
	-8.751538e-001f,
	1.679772e+000f,
	5.345784e+001f,
	3.054520e+001f,
	// albedo 1, turbidity 10
	8.640731e-001f,
	9.858301e-001f,
	1.854956e+000f,
	-6.798097e+000f,
	5.936468e+001f,
	3.110255e+001f,
};



float* datasetsXYZ[] =
{
	datasetXYZ1,
	datasetXYZ2,
	datasetXYZ3
};

float* datasetsXYZRad[] =
{
	datasetXYZRad1,
	datasetXYZRad2,
	datasetXYZRad3
};
